from typing import Optional
import uuid
import av
import sys

import pixeltable.env as env
import pixeltable.func as func
import pixeltable.type_system as ts


_format_defaults = { # format -> (codec, ext)
    'wav': ('pcm_s16le', 'wav'),
    'mp3': ('libmp3lame', 'mp3'),
    'flac': ('flac', 'flac'),
    #'mp4': ('aac', 'm4a'),
}

# for mp4:
# - extract_audio() fails with "Application provided invalid, non monotonically increasing dts to muxer in stream 0: 1146 >= 290"
# - chatgpt suggests this can be fixed in the following manner
#     for packet in container.demux(audio_stream):
#         packet.pts = None  # Reset the PTS and DTS to allow FFmpeg to set them automatically
#         packet.dts = None
#         for frame in packet.decode():
#             frame.pts = None
#             for packet in output_stream.encode(frame):
#                 output_container.mux(packet)
#
#     # Flush remaining packets
#     for packet in output_stream.encode():
#         output_container.mux(packet)


_extract_audio_param_types = [
    ts.VideoType(nullable=False),
    ts.IntType(nullable=False),
    ts.StringType(nullable=False),
    ts.StringType(nullable=False)
]
@func.udf(return_type=ts.AudioType(nullable=True), param_types=_extract_audio_param_types)
def extract_audio(
        video_path: str, stream_idx: int = 0, format: str = 'wav', codec: Optional[str] = None
) -> Optional[str]:
    """Extract an audio stream from a video file, save it as a media file and return its path"""
    if format not in _format_defaults:
        raise ValueError(f'extract_audio(): unsupported audio format: {format}')
    default_codec, ext = _format_defaults[format]

    with av.open(video_path) as container:
        if len(container.streams.audio) <= stream_idx:
            return None
        audio_stream = container.streams.audio[stream_idx]
        # create this in our tmp directory, so it'll get cleaned up if it's being generated as part of a query
        output_filename = str(env.Env.get().tmp_dir / f"{uuid.uuid4()}.{ext}")

        with av.open(output_filename, "w", format=format) as output_container:
            output_stream = output_container.add_stream(codec or default_codec)
            for packet in container.demux(audio_stream):
                for frame in packet.decode():
                    output_container.mux(output_stream.encode(frame))

        return output_filename



from typing import Optional
from types import ModuleType
import importlib
import inspect


def resolve_symbol(symbol_path: str) -> object:
    path_elems = symbol_path.split('.')
    module: Optional[ModuleType] = None
    if path_elems[0:2] == ['pixeltable', 'functions'] and len(path_elems) > 2:
        # if this is a pixeltable.functions submodule, it cannot be resolved via pixeltable.functions;
        # try to import the submodule directly
        submodule_path = '.'.join(path_elems[0:3])
        try:
            module = importlib.import_module(submodule_path)
            path_elems = path_elems[3:]
        except ModuleNotFoundError:
            pass
    if module is None:
        module = importlib.import_module(path_elems[0])
        path_elems = path_elems[1:]
    obj = module
    for el in path_elems:
        obj = getattr(obj, el)
    return obj

def get_caller_module_path() -> str:
    """Return the module path of our caller's caller"""
    stack = inspect.stack()
    try:
        caller_frame = stack[2].frame
        module_path = caller_frame.f_globals['__name__']
    finally:
        # remove references to stack frames to avoid reference cycles
        del stack
    return module_path



from __future__ import annotations
from typing import List, Iterator, Optional
import logging

import pixeltable.exprs as exprs
import pixeltable.catalog as catalog
from pixeltable.utils.media_store import MediaStore


_logger = logging.getLogger('pixeltable')

class DataRowBatch:
    """Set of DataRows, indexed by rowid.

    Contains the metadata needed to initialize DataRows.
    """
    def __init__(self, tbl: catalog.TableVersion, row_builder: exprs.RowBuilder, len: int = 0):
        self.tbl_id = tbl.id
        self.tbl_version = tbl.version
        self.row_builder = row_builder
        self.img_slot_idxs = [e.slot_idx for e in row_builder.unique_exprs if e.col_type.is_image_type()]
        # non-image media slots
        self.media_slot_idxs = [
            e.slot_idx for e in row_builder.unique_exprs
            if e.col_type.is_media_type() and not e.col_type.is_image_type()
        ]
        self.array_slot_idxs = [e.slot_idx for e in row_builder.unique_exprs if e.col_type.is_array_type()]
        self.rows = [
            exprs.DataRow(row_builder.num_materialized, self.img_slot_idxs, self.media_slot_idxs, self.array_slot_idxs)
            for _ in range(len)
        ]

    def add_row(self, row: Optional[exprs.DataRow] = None) -> exprs.DataRow:
        if row is None:
            row = exprs.DataRow(
                self.row_builder.num_materialized, self.img_slot_idxs, self.media_slot_idxs, self.array_slot_idxs)
        self.rows.append(row)
        return row

    def pop_row(self) -> exprs.DataRow:
        return self.rows.pop()

    def set_row_ids(self, row_ids: List[int]) -> None:
        """Sets pks for rows in batch"""
        assert len(row_ids) == len(self.rows)
        for row, row_id in zip(self.rows, row_ids):
            row.set_pk((row_id, self.tbl_version))

    def __len__(self) -> int:
        return len(self.rows)

    def __getitem__(self, index: object) -> exprs.DataRow:
        return self.rows[index]

    def flush_imgs(
            self, idx_range: Optional[slice] = None, stored_img_info: Optional[List[exprs.ColumnSlotIdx]] = None,
            flushed_slot_idxs: Optional[List[int]] = None
    ) -> None:
        """Flushes images in the given range of rows."""
        if stored_img_info is None:
            stored_img_info = []
        if flushed_slot_idxs is None:
            flushed_slot_idxs = []
        if len(stored_img_info) == 0 and len(flushed_slot_idxs) == 0:
            return
        if idx_range is None:
            idx_range = slice(0, len(self.rows))
        for row in self.rows[idx_range]:
            for info in stored_img_info:
                filepath = str(MediaStore.prepare_media_path(self.tbl_id, info.col.id, self.tbl_version))
                row.flush_img(info.slot_idx, filepath)
            for slot_idx in flushed_slot_idxs:
                row.flush_img(slot_idx)
        #_logger.debug(
            #f'flushed images in range {idx_range}: slot_idxs={flushed_slot_idxs} stored_img_info={stored_img_info}')

    def __iter__(self) -> Iterator[exprs.DataRow]:
        return DataRowBatchIterator(self)


class DataRowBatchIterator:
    """
    Iterator over a DataRowBatch.
    """
    def __init__(self, batch: DataRowBatch):
        self.row_batch = batch
        self.index = 0

    def __next__(self) -> exprs.DataRow:
        if self.index >= len(self.row_batch.rows):
            raise StopIteration
        row = self.row_batch.rows[self.index]
        self.index += 1
        return row




from .catalog import Catalog
from .column import Column
from .table_version_path import TableVersionPath
from .table_version import TableVersion
from .schema_object import SchemaObject
from .named_function import NamedFunction
from .dir import Dir
from .table import Table
from .insertable_table import InsertableTable
from .view import View
from .path import Path
from .path_dict import PathDict
from .globals import is_valid_identifier, is_valid_path



import dataclasses
import importlib
import os
import pkgutil
from typing import Callable, Dict

import sqlalchemy as sql
import sqlalchemy.orm as orm

from .schema import SystemInfo, SystemInfoMd

# current version of the metadata; this is incremented whenever the metadata schema changes
VERSION = 12


def create_system_info(engine: sql.engine.Engine) -> None:
    """Create the systemmetadata record"""
    system_md = SystemInfoMd(schema_version=VERSION)
    record = SystemInfo(md=dataclasses.asdict(system_md))
    with orm.Session(engine, future=True) as session:
        session.add(record)
        session.flush()
        session.commit()

# conversion functions for upgrading the metadata schema from one version to the following
# key: old schema version
converter_cbs: Dict[int, Callable[[sql.engine.Engine], None]] = {}

def register_converter(version: int, cb: Callable[[sql.engine.Engine], None]) -> None:
    global converter_cbs
    converter_cbs[version] = cb

# load all converter modules
for _, modname, _ in pkgutil.iter_modules([os.path.dirname(__file__) + '/converters']):
    importlib.import_module('pixeltable.metadata.converters.' + modname)

def upgrade_md(engine: sql.engine.Engine) -> None:
    """Upgrade the metadata schema to the current version"""
    with orm.Session(engine, future=True) as session:
        system_info = session.query(SystemInfo).one().md
        md_version = system_info['schema_version']
        if md_version == VERSION:
                return
        while md_version < VERSION:
            if md_version not in converter_cbs:
                raise RuntimeError(f'No metadata converter for version {md_version}')
            print(f'Converting metadata from version {md_version} to {md_version + 1}')
            converter_cbs[md_version](engine)
            md_version += 1
        # update system info
        conn = session.connection()
        system_info_md = SystemInfoMd(schema_version=VERSION)
        conn.execute(SystemInfo.__table__.update().values(md=dataclasses.asdict(system_info_md)))
        session.commit()



from __future__ import annotations

import dataclasses
import importlib
import inspect
import logging
import time
from typing import Optional, List, Dict, Any, Tuple, Type, Set
from uuid import UUID

import sqlalchemy as sql
import sqlalchemy.orm as orm

import pixeltable
import pixeltable.func as func
from pixeltable import exceptions as excs
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator
from pixeltable.metadata import schema
from pixeltable.utils.filecache import FileCache
from pixeltable.utils.media_store import MediaStore
from .column import Column
from .globals import UpdateStatus, POS_COLUMN_NAME, is_valid_identifier

_logger = logging.getLogger('pixeltable')

class TableVersion:
    """
    TableVersion represents a particular version of a table/view along with its store table:
    - the version can be mutable or a snapshot
    - tables and their recursive views form a tree, and a mutable TableVersion also records its own
      mutable views in order to propagate updates
    - each view TableVersion records its base:
      * the base is correct only for mutable views (snapshot versions form a DAG, not a tree)
      * the base is useful for getting access to the StoreTable and the base id
      * TODO: create a separate hierarchy of objects that records the version-independent tree of tables/views, and
        have TableVersions reference those
    - mutable TableVersions record their TableVersionPath, which is needed for expr evaluation in updates
    """

    def __init__(
            self, id: UUID, tbl_md: schema.TableMd, version: int, schema_version_md: schema.TableSchemaVersionMd,
            base: Optional[TableVersion] = None, base_path: Optional['pixeltable.catalog.TableVersionPath'] = None,
            is_snapshot: Optional[bool] = None
    ):
        # only one of base and base_path can be non-None
        assert base is None or base_path is None
        self.id = id
        self.name = tbl_md.name
        self.version = version
        self.comment = schema_version_md.comment
        self.num_retained_versions = schema_version_md.num_retained_versions
        self.schema_version = schema_version_md.schema_version
        self.view_md = tbl_md.view_md  # save this as-is, it's needed for _create_md()
        is_view = tbl_md.view_md is not None
        self.is_snapshot = (is_view and tbl_md.view_md.is_snapshot) or bool(is_snapshot)
        # a mutable TableVersion doesn't have a static version
        self.effective_version = self.version if self.is_snapshot else None

        # mutable tables need their TableVersionPath for expr eval during updates
        from .table_version_path import TableVersionPath
        if self.is_snapshot:
            self.path = None
        else:
            self.path = TableVersionPath(self, base=base_path) if base_path is not None else TableVersionPath(self)

        self.base = base_path.tbl_version if base_path is not None else base
        if self.is_snapshot:
            self.next_col_id = -1
            self.next_rowid = -1
        else:
            assert tbl_md.current_version == self.version
            self.next_col_id = tbl_md.next_col_id
            self.next_rowid = tbl_md.next_row_id
        self.column_history = tbl_md.column_history

        # view-specific initialization
        from pixeltable import exprs
        predicate_dict = None if not is_view or tbl_md.view_md.predicate is None else tbl_md.view_md.predicate
        self.predicate = exprs.Expr.from_dict(predicate_dict) if predicate_dict is not None else None
        self.mutable_views: List[TableVersion] = []  # targets for update propagation
        if self.base is not None and not self.base.is_snapshot and not self.is_snapshot:
            self.base.mutable_views.append(self)

        # component view-specific initialization
        self.iterator_cls: Optional[Type[ComponentIterator]] = None
        self.iterator_args: Optional[exprs.InlineDict] = None
        self.num_iterator_cols = 0
        if is_view and tbl_md.view_md.iterator_class_fqn is not None:
            module_name, class_name = tbl_md.view_md.iterator_class_fqn.rsplit('.', 1)
            module = importlib.import_module(module_name)
            self.iterator_cls = getattr(module, class_name)
            self.iterator_args = exprs.Expr.from_dict(tbl_md.view_md.iterator_args)
            assert isinstance(self.iterator_args, exprs.InlineDict)
            output_schema, _ = self.iterator_cls.output_schema(**self.iterator_args.to_dict())
            self.num_iterator_cols = len(output_schema)
            assert tbl_md.view_md.iterator_args is not None

        # register this table version now so that it's available when we're re-creating value exprs
        import pixeltable.catalog as catalog
        cat = catalog.Catalog.get()
        cat.tbl_versions[(self.id, self.effective_version)] = self

        # do this after we determined whether we're a component view, and before we create the store table
        self._init_schema(schema_version_md)

    def __hash__(self) -> int:
        return hash(self.id)

    def create_snapshot_copy(self) -> TableVersion:
        """Create a snapshot copy of this TableVersion"""
        assert not self.is_snapshot
        return TableVersion(
            self.id, self._create_md(), self.version,
            self._create_schema_version_md(preceding_schema_version=0),  # preceding_schema_version: dummy value
            is_snapshot=True, base=self.base)

    @classmethod
    def create(
            cls, session: orm.Session, dir_id: UUID, name: str, cols: List[Column], num_retained_versions: int, comment: str,
            base_path: Optional['pixeltable.catalog.TableVersionPath'] = None, view_md: Optional[schema.ViewMd] = None
    ) -> Tuple[UUID, Optional[TableVersion]]:
        # assign ids
        cols_by_name: Dict[str, Column] = {}
        for pos, col in enumerate(cols):
            col.id = pos
            cols_by_name[col.name] = col
            if col.value_expr is None and col.compute_func is not None:
                cls._create_value_expr(col, base_path)
            if col.is_computed:
                col.check_value_expr()

        ts = time.time()
        # create schema.Table
        column_history = {
            col.id: schema.ColumnHistory(col_id=col.id, schema_version_add=0, schema_version_drop=None)
            for col in cols
        }
        table_md = schema.TableMd(
            name=name, current_version=0, current_schema_version=0,
            next_col_id=len(cols), next_row_id=0, column_history=column_history,
            view_md=view_md)
        tbl_record = schema.Table(dir_id=dir_id, md=dataclasses.asdict(table_md))
        session.add(tbl_record)
        session.flush()  # sets tbl_record.id
        assert tbl_record.id is not None

        # create schema.TableVersion
        table_version_md = schema.TableVersionMd(created_at=ts, version=0, schema_version=0)
        tbl_version_record = schema.TableVersion(
            tbl_id=tbl_record.id, version=0, md=dataclasses.asdict(table_version_md))
        session.add(tbl_version_record)

        # create schema.TableSchemaVersion
        column_md: Dict[int, schema.SchemaColumn] = {}
        for pos, col in enumerate(cols):
            # Column.dependent_cols for existing cols is wrong at this point, but init() will set it correctly
            value_expr_dict = col.value_expr.as_dict() if col.value_expr is not None else None
            column_md[col.id] = schema.SchemaColumn(
                pos=pos, name=col.name, col_type=col.col_type.as_dict(),
                is_pk=col.primary_key, value_expr=value_expr_dict, stored=col.stored, is_indexed=col.is_indexed)

        schema_version_md = schema.TableSchemaVersionMd(
            schema_version=0, preceding_schema_version=None, columns=column_md,
            num_retained_versions=num_retained_versions, comment=comment)
        schema_version_record = schema.TableSchemaVersion(
            tbl_id=tbl_record.id, schema_version=0, md=dataclasses.asdict(schema_version_md))
        session.add(schema_version_record)

        # if this is purely a snapshot (it doesn't require any additional storage for columns and it # doesn't have a
        # predicate to apply at runtime), we don't create a physical table and simply use the base's table version path
        if view_md is not None and view_md.is_snapshot and view_md.predicate is None and len(cols) == 0:
            return tbl_record.id, None

        assert (base_path is not None) == (view_md is not None)
        base = base_path.tbl_version if base_path is not None and view_md.is_snapshot else None
        base_path = base_path if base_path is not None and not view_md.is_snapshot else None
        tbl_version = cls(tbl_record.id, table_md, 0, schema_version_md, base=base, base_path=base_path)
        tbl_version.store_tbl.create(session.connection())
        # TODO: create pgvector indices
        return tbl_record.id, tbl_version

    @classmethod
    def delete_md(cls, tbl_id: UUID, conn: sql.Connection) -> None:
        conn.execute(
            sql.delete(schema.TableSchemaVersion.__table__).where(schema.TableSchemaVersion.tbl_id == tbl_id))
        conn.execute(
            sql.delete(schema.TableVersion.__table__).where(schema.TableVersion.tbl_id == tbl_id))
        conn.execute(sql.delete(schema.Table.__table__).where(schema.Table.id == tbl_id))

    def drop(self) -> None:
        with Env.get().engine.begin() as conn:
            # delete this table and all associated data
            MediaStore.delete(self.id)
            FileCache.get().clear(tbl_id=self.id)
            self.delete_md(self.id, conn)
            self.store_tbl.drop(conn)

        # de-register table version from catalog
        from .catalog import Catalog
        cat = Catalog.get()
        del cat.tbl_versions[(self.id, self.effective_version)]
        # TODO: remove from tbl_dependents

    def _init_schema(self, schema_version_md: schema.TableSchemaVersionMd) -> None:
        """Initialize self.cols as well as self.store_tbl"""
        self.cols = [Column.from_md(col_id, col_md, self) for col_id, col_md in schema_version_md.columns.items()]
        self.cols_by_name = {col.name: col for col in self.cols}
        self.cols_by_id = {col.id: col for col in self.cols}

        # make sure to traverse columns ordered by position = order in which cols were created;
        # this guarantees that references always point backwards
        from pixeltable import exprs
        for col, col_md in zip(self.cols, schema_version_md.columns.values()):
            col.tbl = self
            if col_md.value_expr is not None:
                col.value_expr = exprs.Expr.from_dict(col_md.value_expr)
                self._record_value_expr(col)

        # create the sqlalchemy schema; do this after instantiating columns, in order to determine whether they
        # need to record errors
        from pixeltable.store import StoreBase, StoreTable, StoreView, StoreComponentView
        if self.is_component_view():
            self.store_tbl: StoreBase = StoreComponentView(self)
        elif self.is_view():
            self.store_tbl: StoreBase = StoreView(self)
        else:
            self.store_tbl: StoreBase = StoreTable(self)

    def _update_md(
            self, ts: float, preceding_schema_version: Optional[int], conn: sql.engine.Connection) -> None:
        """Update all recorded metadata in response to a data or schema change.
        Args:
            ts: timestamp of the change
            preceding_schema_version: last schema version if schema change, else None
        """
        conn.execute(
            sql.update(schema.Table.__table__)
                .values({schema.Table.md: dataclasses.asdict(self._create_md())})
                .where(schema.Table.id == self.id))
        version_md = self._create_version_md(ts)
        conn.execute(
            sql.insert(schema.TableVersion.__table__)
                .values(tbl_id=self.id, version=self.version, md=dataclasses.asdict(version_md)))
        if preceding_schema_version is not None:
            schema_version_md = self._create_schema_version_md(preceding_schema_version)
            conn.execute(
                sql.insert(schema.TableSchemaVersion.__table__)
                .values(
                    tbl_id=self.id, schema_version=self.schema_version,
                    md=dataclasses.asdict(schema_version_md)))

    def add_column(self, col: Column, print_stats: bool = False) -> UpdateStatus:
        """Adds a column to the table.
        """
        assert not self.is_snapshot
        assert is_valid_identifier(col.name)
        assert col.stored is not None
        assert col.name not in self.cols_by_name
        col.tbl = self
        col.id = self.next_col_id
        self.next_col_id += 1

        if col.compute_func is not None:
            # create value_expr from compute_func
            self._create_value_expr(col, self.path)
        if col.value_expr is not None:
            col.check_value_expr()
            self._record_value_expr(col)

        row_count = self.store_tbl.count()
        if row_count > 0 and not col.col_type.nullable and not col.is_computed:
            raise excs.Error(f'Cannot add non-nullable column "{col.name}" to table {self.name} with existing rows')

        # we're creating a new schema version
        ts = time.time()
        self.version += 1
        preceding_schema_version = self.schema_version
        self.schema_version = self.version

        self.cols.append(col)
        self.cols_by_name[col.name] = col
        self.cols_by_id[col.id] = col
        self.column_history[col.id] = schema.ColumnHistory(col.id, self.schema_version, None)

        with Env.get().engine.begin() as conn:
            self._update_md(ts, preceding_schema_version, conn)
            _logger.info(f'Added column {col.name} to table {self.name}, new version: {self.version}')
            if col.is_stored:
                self.store_tbl.add_column(col, conn)

        print(f'Added column `{col.name}` to table `{self.name}`.')
        if row_count == 0:
            return UpdateStatus()
        if (not col.is_computed or not col.is_stored) and not col.is_indexed:
            return UpdateStatus(num_rows=row_count)
        # compute values for the existing rows and compute embeddings, if this column is indexed;
        # for some reason, it's not possible to run the following updates in the same transaction as the one
        # that we just used to create the metadata (sqlalchemy hangs when exec() tries to run the query)
        from pixeltable.plan import Planner
        plan, value_expr_slot_idx, embedding_slot_idx = Planner.create_add_column_plan(self.path, col)
        plan.ctx.num_rows = row_count
        # TODO: create pgvector index, if col is indexed

        try:
            # TODO: do this in the same transaction as the metadata update
            with Env.get().engine.begin() as conn:
                plan.ctx.conn = conn
                plan.open()
                num_excs = self.store_tbl.load_column(col, plan, value_expr_slot_idx, embedding_slot_idx, conn)
        except sql.exc.DBAPIError as e:
            self.drop_column(col.name)
            raise excs.Error(f'Error during SQL execution:\n{e}')
        finally:
            plan.close()

        msg = f'Added {row_count} column value{"" if row_count == 1 else "s"} with {num_excs} error{"" if num_excs == 1 else "s"}.'
        print(msg)
        _logger.info(f'Column {col.name}: {msg}')
        if print_stats:
            plan.ctx.profile.print(num_rows=row_count)
        return UpdateStatus(
            num_rows=row_count, num_computed_values=row_count, num_excs=num_excs,
            cols_with_excs=[f'{self.name}.{col.name}'] if num_excs > 0 else [])

    def drop_column(self, name: str) -> None:
        """Drop a column from the table.
        """
        assert not self.is_snapshot
        if name not in self.cols_by_name:
            raise excs.Error(f'Unknown column: {name}')
        col = self.cols_by_name[name]
        if len(col.dependent_cols) > 0:
            raise excs.Error(
                f'Cannot drop column {name} because the following columns depend on it:\n',
                f'{", ".join([c.name for c in col.dependent_cols])}')

        if col.value_expr is not None:
            # update Column.dependent_cols
            for c in self.cols:
                if c == col:
                    break
                c.dependent_cols.discard(col)

        # we're creating a new schema version
        ts = time.time()
        self.version += 1
        preceding_schema_version = self.schema_version
        self.schema_version = self.version

        self.cols.remove(col)
        del self.cols_by_name[name]
        del self.cols_by_id[col.id]
        self.column_history[col.id].schema_version_drop = self.schema_version

        with Env.get().engine.begin() as conn:
            self._update_md(ts, preceding_schema_version, conn)
        if col.is_stored:
            self.store_tbl.drop_column()
        _logger.info(f'Dropped column {name} from table {self.name}, new version: {self.version}')

    def rename_column(self, old_name: str, new_name: str) -> None:
        """Rename a column.
        """
        assert not self.is_snapshot
        if old_name not in self.cols_by_name:
            raise excs.Error(f'Unknown column: {old_name}')
        if not is_valid_identifier(new_name):
            raise excs.Error(f"Invalid column name: '{new_name}'")
        if new_name in self.cols_by_name:
            raise excs.Error(f'Column {new_name} already exists')
        col = self.cols_by_name[old_name]
        del self.cols_by_name[old_name]
        col.name = new_name
        self.cols_by_name[new_name] = col

        # we're creating a new schema version
        ts = time.time()
        self.version += 1
        preceding_schema_version = self.schema_version
        self.schema_version = self.version

        with Env.get().engine.begin() as conn:
            self._update_md(ts, preceding_schema_version, conn)
        _logger.info(f'Renamed column {old_name} to {new_name} in table {self.name}, new version: {self.version}')

    def set_comment(self, new_comment: Optional[str]):
        _logger.info(f'[{self.name}] Updating comment: {new_comment}')
        self.comment = new_comment
        self._commit_new_schema_version()

    def set_num_retained_versions(self, new_num_retained_versions: int):
        _logger.info(f'[{self.name}] Updating num_retained_versions: {new_num_retained_versions} (was {self.num_retained_versions})')
        self.num_retained_versions = new_num_retained_versions
        self._commit_new_schema_version()

    def _commit_new_schema_version(self):
        # we're creating a new schema version
        ts = time.time()
        self.version += 1
        preceding_schema_version = self.schema_version
        self.schema_version = self.version
        with Env.get().engine.begin() as conn:
            self._update_md(ts, preceding_schema_version, conn)
        _logger.info(f'[{self.name}] Updating table schema to version: {self.version}')

    def insert(
            self, rows: List[Dict[str, Any]], print_stats: bool = False, fail_on_exception : bool = True
    ) -> UpdateStatus:
        """Insert rows into this table.
        """
        assert self.is_insertable()
        from pixeltable.plan import Planner
        plan = Planner.create_insert_plan(self, rows, ignore_errors=not fail_on_exception)
        ts = time.time()
        with Env.get().engine.begin() as conn:
            return self._insert(plan, conn, ts, print_stats)

    def _insert(
            self, exec_plan: exec.ExecNode, conn: sql.engine.Connection, ts: float, print_stats: bool = False,
    ) -> UpdateStatus:
        """Insert rows produced by exec_plan and propagate to views"""
        # we're creating a new version
        self.version += 1
        result = UpdateStatus()
        num_rows, num_excs, cols_with_excs = self.store_tbl.insert_rows(exec_plan, conn, v_min=self.version)
        self.next_rowid = num_rows
        result.num_rows = num_rows
        result.num_excs = num_excs
        result.num_computed_values += exec_plan.ctx.num_computed_exprs * num_rows
        result.cols_with_excs = [f'{self.name}.{self.cols_by_id[cid].name}' for cid in cols_with_excs]
        self._update_md(ts, None, conn)

        # update views
        for view in self.mutable_views:
            from pixeltable.plan import Planner
            plan, _ = Planner.create_view_load_plan(view.path, propagates_insert=True)
            status = view._insert(plan, conn, ts, print_stats)
            result.num_rows += status.num_rows
            result.num_excs += status.num_excs
            result.num_computed_values += status.num_computed_values
            result.cols_with_excs += status.cols_with_excs

        result.cols_with_excs = list(dict.fromkeys(result.cols_with_excs).keys())  # remove duplicates
        if print_stats:
            plan.ctx.profile.print(num_rows=num_rows)
        _logger.info(f'TableVersion {self.name}: new version {self.version}')
        return result

    def update(
            self, update_targets: Optional[List[Tuple[Column, 'pixeltable.exprs.Expr']]] = None,
            where_clause: Optional['pixeltable.exprs.Predicate'] = None, cascade: bool = True
    ) -> UpdateStatus:
        """Update rows in this table.
        Args:
            update_targets: a list of (column, value) pairs specifying the columns to update and their new values.
            where_clause: a Predicate to filter rows to update.
            cascade: if True, also update all computed columns that transitively depend on the updated columns,
                including within views.
        """
        if update_targets is None:
            update_targets = []
        assert not self.is_snapshot
        from pixeltable.plan import Planner
        plan, updated_cols, recomputed_cols = \
            Planner.create_update_plan(self.path, update_targets, [], where_clause, cascade)
        with Env.get().engine.begin() as conn:
            ts = time.time()
            result = self._update(
                plan, where_clause.sql_expr() if where_clause is not None else None, recomputed_cols,
                base_versions=[], conn=conn, ts=ts, cascade=cascade)
            result.updated_cols = updated_cols
            return result

    def _update(
            self, plan: Optional[exec.ExecNode], where_clause: Optional[sql.ClauseElement],
            recomputed_view_cols: List[Column], base_versions: List[Optional[int]], conn: sql.engine.Connection,
            ts: float, cascade: bool
    ) -> UpdateStatus:
        result = UpdateStatus()
        if plan is not None:
            # we're creating a new version
            self.version += 1
            result.num_rows, result.num_excs, cols_with_excs = \
                self.store_tbl.insert_rows(plan, conn, v_min=self.version)
            result.cols_with_excs = [f'{self.name}.{self.cols_by_id[cid].name}' for cid in cols_with_excs]
            self.store_tbl.delete_rows(
                self.version, base_versions=base_versions, match_on_vmin=True, where_clause=where_clause, conn=conn)
            self._update_md(ts, None, conn)

        if cascade:
            base_versions = [None if plan is None else self.version] + base_versions  # don't update in place
            # propagate to views
            for view in self.mutable_views:
                recomputed_cols = [col for col in recomputed_view_cols if col.tbl is view]
                plan: Optional[exec.ExecNode] = None
                if len(recomputed_cols) > 0:
                    from pixeltable.plan import Planner
                    plan = Planner.create_view_update_plan(view.path, recompute_targets=recomputed_cols)
                status = view._update(
                    plan, None, recomputed_view_cols, base_versions=base_versions, conn=conn, ts=ts, cascade=True)
                result.num_rows += status.num_rows
                result.num_excs += status.num_excs
                result.cols_with_excs += status.cols_with_excs

        result.cols_with_excs = list(dict.fromkeys(result.cols_with_excs).keys())  # remove duplicates
        return result

    def delete(self, where: Optional['pixeltable.exprs.Predicate'] = None) -> UpdateStatus:
        """Delete rows in this table.
        Args:
            where: a Predicate to filter rows to delete.
        """
        assert self.is_insertable()
        from pixeltable.plan import Planner
        analysis_info = Planner.analyze(self, where)
        ts = time.time()
        with Env.get().engine.begin() as conn:
            num_rows = self._delete(analysis_info.sql_where_clause, base_versions=[], conn=conn, ts=ts)

        status = UpdateStatus(num_rows=num_rows)
        return status

    def _delete(
            self, where: Optional['pixeltable.exprs.Predicate'], base_versions: List[Optional[int]],
            conn: sql.engine.Connection, ts: float) -> int:
        """Delete rows in this table and propagate to views.
        Args:
            where: a Predicate to filter rows to delete.
        Returns:
            number of deleted rows
        """
        sql_where_clause = where.sql_expr() if where is not None else None
        num_rows = self.store_tbl.delete_rows(
            self.version + 1, base_versions=base_versions, match_on_vmin=False, where_clause=sql_where_clause,
            conn=conn)
        if num_rows > 0:
            # we're creating a new version
            self.version += 1
            self._update_md(ts, None, conn)
        else:
            pass
        for view in self.mutable_views:
            num_rows += view._delete(where=None, base_versions=[self.version] + base_versions, conn=conn, ts=ts)
        return num_rows

    def revert(self) -> None:
        """Reverts the table to the previous version.
        """
        assert not self.is_snapshot
        if self.version == 0:
            raise excs.Error('Cannot revert version 0')
        with orm.Session(Env.get().engine, future=True) as session:
            self._revert(session)
            session.commit()

    def _revert(self, session: orm.Session) -> None:
        """Reverts this table version and propagates to views"""
        conn = session.connection()
        # make sure we don't have a snapshot referencing this version
        # (unclear how to express this with sqlalchemy)
        query = (
            f"select ts.dir_id, ts.md->'name' "
            f"from {schema.Table.__tablename__} ts "
            f"cross join lateral jsonb_path_query(md, '$.view_md.base_versions[*]') as tbl_version "
            f"where tbl_version->>0 = '{self.id.hex}' and (tbl_version->>1)::int = {self.version}"
        )
        result = list(conn.execute(sql.text(query)))
        if len(result) > 0:
            names = [row[1] for row in result]
            raise excs.Error((
                f'Current version is needed for {len(result)} snapshot{"s" if len(result) > 1 else ""} '
                f'({", ".join(names)})'
            ))

        conn = session.connection()
        # delete newly-added data
        MediaStore.delete(self.id, version=self.version)
        conn.execute(sql.delete(self.store_tbl.sa_tbl).where(self.store_tbl.sa_tbl.c.v_min == self.version))
        # revert new deletions
        conn.execute(
            sql.update(self.store_tbl.sa_tbl) \
                .values({self.store_tbl.sa_tbl.c.v_max: schema.Table.MAX_VERSION})
                .where(self.store_tbl.sa_tbl.c.v_max == self.version))

        if self.version == self.schema_version:
            # the current version involved a schema change:
            # if the schema change was to add a column, we now need to drop it
            added_col_ids = [
                col_history.col_id for col_history in self.column_history.values()
                if col_history.schema_version_add == self.schema_version
            ]
            assert len(added_col_ids) <= 1
            added_col: Optional[Column] = None
            if len(added_col_ids) == 1:
                added_col_id = added_col_ids[0]
                # drop this newly-added column and its ColumnHistory record
                c = self.cols_by_id[added_col_id]
                if c.is_stored:
                    added_col = c
                del self.column_history[c.id]

            # we need to determine the preceding schema version and reload the schema
            schema_version_md_dict = session.query(schema.TableSchemaVersion.md) \
                .where(schema.TableSchemaVersion.tbl_id == self.id) \
                .where(schema.TableSchemaVersion.schema_version == self.schema_version) \
                .scalar()
            preceding_schema_version = schema_version_md_dict['preceding_schema_version']
            preceding_schema_version_md_dict = session.query(schema.TableSchemaVersion.md) \
                .where(schema.TableSchemaVersion.tbl_id == self.id) \
                .where(schema.TableSchemaVersion.schema_version == preceding_schema_version) \
                .scalar()
            preceding_schema_version_md = schema.md_from_dict(
                schema.TableSchemaVersionMd, preceding_schema_version_md_dict)
            self._init_schema(preceding_schema_version_md)

            # physically drop the column, but only after we have re-created the schema
            if added_col is not None:
                self.store_tbl.drop_column(added_col, conn)

            conn.execute(
                sql.delete(schema.TableSchemaVersion.__table__)
                    .where(schema.TableSchemaVersion.tbl_id == self.id)
                    .where(schema.TableSchemaVersion.schema_version == self.schema_version))
            self.schema_version = preceding_schema_version
            self.comment = preceding_schema_version_md.comment
            self.num_retained_versions = preceding_schema_version_md.num_retained_versions

        conn.execute(
            sql.delete(schema.TableVersion.__table__)
                .where(schema.TableVersion.tbl_id == self.id)
                .where(schema.TableVersion.version == self.version)
        )
        self.version -= 1
        conn.execute(
            sql.update(schema.Table.__table__)
                .values({schema.Table.md: dataclasses.asdict(self._create_md())})
                .where(schema.Table.id == self.id))

        # propagate to views
        for view in self.mutable_views:
            view._revert(session)
        _logger.info(f'TableVersion {self.name}: reverted to version {self.version}')

    def is_view(self) -> bool:
        return self.base is not None

    def is_component_view(self) -> bool:
        return self.iterator_cls is not None

    def is_insertable(self) -> bool:
        """Returns True if this corresponds to an InsertableTable"""
        return not self.is_snapshot and not self.is_view()

    def is_iterator_column(self, col: Column) -> bool:
        """Returns True if col is produced by an iterator"""
        # the iterator columns directly follow the pos column
        return self.is_component_view() and col.id > 0 and col.id < self.num_iterator_cols + 1

    def is_system_column(self, col: Column) -> bool:
        """Return True if column was created by Pixeltable"""
        if col.name == POS_COLUMN_NAME and self.is_component_view():
            return True
        return False

    def user_columns(self) -> List[Column]:
        """Return all non-system columns"""
        return [c for c in self.cols if not self.is_system_column(c)]

    def get_required_col_names(self) -> List[str]:
        """Return the names of all columns for which values must be specified in insert()"""
        assert not self.is_view()
        names = [c.name for c in self.cols if not c.is_computed and not c.col_type.nullable]
        return names

    def get_computed_col_names(self) -> List[str]:
        """Return the names of all computed columns"""
        names = [c.name for c in self.cols if c.is_computed]
        return names

    @classmethod
    def _create_value_expr(cls, col: Column, path: 'TableVersionPath') -> None:
        """
        Create col.value_expr, given col.compute_func.
        Interprets compute_func's parameters to be references to columns and construct ColumnRefs as args.
        Does not update Column.dependent_cols.
        """
        assert col.value_expr is None
        assert col.compute_func is not None
        from pixeltable import exprs
        params = inspect.signature(col.compute_func).parameters
        args: List[exprs.ColumnRef] = []
        for param_name in params:
            param = path.get_column(param_name)
            if param is None:
                raise excs.Error(
                    f'Column {col.name}: Callable parameter refers to an unknown column: {param_name}')
            args.append(exprs.ColumnRef(param))
        fn = func.make_function(
            col.compute_func, return_type=col.col_type, param_types=[arg.col_type for arg in args])
        col.value_expr = fn(*args)

    def _record_value_expr(self, col: Column) -> None:
        """Update Column.dependent_cols for all cols referenced in col.value_expr.
        """
        assert col.value_expr is not None
        from pixeltable.exprs import ColumnRef
        refd_cols = [e.col for e in col.value_expr.subexprs(expr_class=ColumnRef)]
        for refd_col in refd_cols:
            refd_col.dependent_cols.add(col)

    def get_dependent_columns(self, cols: List[Column]) -> Set[Column]:
        """
        Return the set of columns that transitively depend on any of the given ones.
        """
        if len(cols) == 0:
            return []
        result: Set[Column] = set()
        for col in cols:
            result.update(col.dependent_cols)
        result.update(self.get_dependent_columns(result))
        return result

    def num_rowid_columns(self) -> int:
        """Return the number of columns of the rowids, without accessing store_tbl"""
        if self.is_component_view():
            return 1 + self.base.num_rowid_columns()
        return 1

    def _create_md(self) -> schema.TableMd:
        return schema.TableMd(
            name=self.name, current_version=self.version, current_schema_version=self.schema_version,
            next_col_id=self.next_col_id, next_row_id=self.next_rowid, column_history=self.column_history,
            view_md=self.view_md)

    def _create_version_md(self, ts: float) -> schema.TableVersionMd:
        return schema.TableVersionMd(created_at=ts, version=self.version, schema_version=self.schema_version)

    def _create_schema_version_md(self, preceding_schema_version: int) -> schema.TableSchemaVersionMd:
        column_md: Dict[int, schema.SchemaColumn] = {}
        for pos, col in enumerate(self.cols):
            value_expr_dict = col.value_expr.as_dict() if col.value_expr is not None else None
            column_md[col.id] = schema.SchemaColumn(
                pos=pos, name=col.name, col_type=col.col_type.as_dict(),
                is_pk=col.primary_key, value_expr=value_expr_dict, stored=col.stored, is_indexed=col.is_indexed)
        # preceding_schema_version to be set by the caller
        return schema.TableSchemaVersionMd(
            schema_version=self.schema_version, preceding_schema_version=preceding_schema_version,
            columns=column_md, num_retained_versions=self.num_retained_versions, comment=self.comment)



from __future__ import annotations

import abc
import importlib
import inspect
import pixeltable
from typing import Optional, Dict, Any, Tuple

from .globals import resolve_symbol
from .signature import Signature


class Function(abc.ABC):
    """Base class for Pixeltable's function interface.

    A function in Pixeltable is an object that has a signature and implements __call__().
    This base class provides a default serialization mechanism for Function instances provided by Python modules,
    via the member self_path.
    """

    def __init__(self, signature: Signature, py_signature: inspect.Signature, self_path: Optional[str] = None):
        self.signature = signature
        self.py_signature = py_signature
        self.self_path = self_path  # fully-qualified path to self

    @property
    def name(self) -> str:
        assert self.self_path is not None
        return self.self_path.split('.')[-1]

    @property
    def display_name(self) -> str:
        if self.self_path is None:
            return '<anonymous>'
        ptf_prefix = 'pixeltable.functions.'
        if self.self_path.startswith(ptf_prefix):
            return self.self_path[len(ptf_prefix):]
        return self.self_path

    def help_str(self) -> str:
        return self.display_name + str(self.signature)

    def __call__(self, *args: object, **kwargs: object) -> 'pixeltable.exprs.Expr':
        from pixeltable import exprs
        bound_args = self.py_signature.bind(*args, **kwargs)
        self.validate_call(bound_args.arguments)
        return exprs.FunctionCall(self, bound_args.arguments)

    def validate_call(self, bound_args: Dict[str, Any]) -> None:
        """Override this to do custom validation of the arguments"""
        pass

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.self_path == other.self_path

    def source(self) -> None:
        """Print source code"""
        print('source not available')

    def as_dict(self) -> Dict:
        """
        Return a serialized reference to the instance that can be passed to json.dumps() and converted back
        to an instance with from_dict().
        Subclasses can override _as_dict().
        """
        classpath = f'{self.__class__.__module__}.{self.__class__.__qualname__}'
        return {'_classpath': classpath, **self._as_dict()}

    def _as_dict(self) -> Dict:
        """Default serialization: store the path to self (which includes the module path)"""
        assert self.self_path is not None
        return {'path': self.self_path}

    @classmethod
    def from_dict(cls, d: Dict) -> Function:
        """
        Turn dict that was produced by calling as_dict() into an instance of the correct Function subclass.
        """
        assert '_classpath' in d
        module_path, class_name = d['_classpath'].rsplit('.', 1)
        class_module = importlib.import_module(module_path)
        func_class = getattr(class_module, class_name)
        return func_class._from_dict(d)

    @classmethod
    def _from_dict(cls, d: Dict) -> Function:
        """Default deserialization: load the symbol indicated by the stored symbol_path"""
        assert 'path' in d and d['path'] is not None
        instance = resolve_symbol(d['path'])
        assert isinstance(instance, Function)
        return instance

    def to_store(self) -> Tuple[Dict, bytes]:
        """
        Serialize the function to a format that can be stored in the Pixeltable store
        Returns:
            - a dict that can be passed to json.dumps()
            - additional binary data
        Only Function subclasses that can be stored need to override this.
        """
        raise NotImplementedError()

    @classmethod
    def from_store(cls, name: Optional[str], md: Dict, binary_obj: bytes) -> Function:
        """
        Create a Function instance from the serialized representation returned by to_store()
        """
        raise NotImplementedError()



from __future__ import annotations

import abc
import hashlib
import importlib
import inspect
import json
import sys
import typing
from itertools import islice
from typing import Union, Optional, List, Callable, Any, Dict, Tuple, Set, Generator, Type
from uuid import UUID

import sqlalchemy as sql

import pixeltable
import pixeltable.catalog as catalog
import pixeltable.exceptions as excs
import pixeltable.type_system as ts
import pixeltable.func as func
from .data_row import DataRow
from .globals import ComparisonOperator, LogicalOperator, LiteralPythonTypes, ArithmeticOperator


class ExprScope:
    """
    Representation of the scope in which an Expr needs to be evaluated. Used to determine nesting of scopes.
    parent is None: outermost scope
    """
    def __init__(self, parent: Optional[ExprScope]):
        self.parent = parent

    def is_contained_in(self, other: ExprScope) -> bool:
        if self == other:
            return True
        if self.parent is None:
            return False
        return self.parent.is_contained_in(other)


_GLOBAL_SCOPE = ExprScope(None)


class Expr(abc.ABC):
    """
    Rules for using state in subclasses:
    - all state except for components and slot_idx is shared between copies of an Expr
    - slot_idx is set during analysis (DataFrame.show())
    - during eval(), components can only be accessed via self.components; any Exprs outside of that won't
      have slot_idx set
    """
    def __init__(self, col_type: ts.ColumnType):
        self.col_type = col_type

        # each instance has an id that is used for equality comparisons
        # - set by the subclass's __init__()
        # - produced by _create_id()
        # - not expected to survive a serialize()/deserialize() roundtrip
        self.id: Optional[int] = None

        # index of the expr's value in the data row:
        # - set for all materialized exprs
        # - -1: not executable
        # - not set for subexprs that don't need to be materialized because the parent can be materialized via SQL
        self.slot_idx = -1
        self.components: List[Expr] = []  # the subexprs that are needed to construct this expr

    def dependencies(self) -> List[Expr]:
        """
        Returns all exprs that need to have been evaluated before eval() can be called on this one.
        """
        return self.components

    def scope(self) -> ExprScope:
        # by default this is the innermost scope of any of our components
        result = _GLOBAL_SCOPE
        for c in self.components:
            c_scope = c.scope()
            if c_scope.is_contained_in(result):
                result = c_scope
        return result

    def bind_rel_paths(self, mapper: Optional['pixeltable.exprs.JsonMapper'] = None) -> None:
        """
        Binds relative JsonPaths to mapper.
        This needs to be done in a separate phase after __init__(), because RelativeJsonPath()(-1) cannot be resolved
        by the immediately containing JsonMapper during initialization.
        """
        for c in self.components:
            c.bind_rel_paths(mapper)

    def default_column_name(self) -> Optional[str]:
        """
        Returns: 
            None if this expression lacks a default name, 
            or a valid identifier (according to catalog.is_valid_identifer) otherwise.
        """
        return None

    def equals(self, other: Expr) -> bool:
        """
        Subclass-specific comparison. Implemented as a function because __eq__() is needed to construct Comparisons.
        """
        if type(self) != type(other):
            return False
        if len(self.components) != len(other.components):
            return False
        for i in range(len(self.components)):
            if not self.components[i].equals(other.components[i]):
                return False
        return self._equals(other)

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        """Returns attribute name/value pairs that are used to construct the instance id.

        Attribute values must be immutable and have str() defined.
        """
        return [('classname', self.__class__.__name__)]

    def _create_id(self) -> int:
        hasher = hashlib.sha256()
        for attr, value in self._id_attrs():
            hasher.update(attr.encode('utf-8'))
            hasher.update(str(value).encode('utf-8'))
        for expr in self.components:
            hasher.update(str(expr.id).encode('utf-8'))
        # truncate to machine's word size
        return int(hasher.hexdigest(), 16) & sys.maxsize

    def __hash__(self) -> int:
        assert self.id is not None
        return self.id

    @classmethod
    def list_equals(cls, a: List[Expr], b: List[Expr]) -> bool:
        if len(a) != len(b):
            return False
        for i in range(len(a)):
            if not a[i].equals(b[i]):
                return False
        return True

    def copy(self) -> Expr:
        """
        Creates a copy that can be evaluated separately: it doesn't share any eval context (slot_idx)
        but shares everything else (catalog objects, etc.)
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.slot_idx = -1
        result.components = [c.copy() for c in self.components]
        return result

    @classmethod
    def copy_list(cls, expr_list: List[Expr]) -> List[Expr]:
        return [e.copy() for e in expr_list]

    def __deepcopy__(self, memo=None) -> Expr:
        # we don't need to create an actual deep copy because all state other than execution state is read-only
        if memo is None:
            memo = {}
        result = self.copy()
        memo[id(self)] = result
        return result

    def substitute(self, old: Expr, new: Expr) -> Expr:
        """
        Replace 'old' with 'new' recursively.
        """
        if self.equals(old):
            return new.copy()
        for i in range(len(self.components)):
            self.components[i] = self.components[i].substitute(old, new)
        return self

    def resolve_computed_cols(self, resolve_cols: Optional[Set[catalog.Column]] = None) -> Expr:
        """
        Recursively replace ColRefs to unstored computed columns with their value exprs.
        Also replaces references to stored computed columns in resolve_cols.
        """
        from .expr_set import ExprSet
        from .column_ref import ColumnRef
        if resolve_cols is None:
            resolve_cols = set()
        result = self
        while True:
            target_col_refs = ExprSet([
                e for e in result.subexprs()
                if isinstance(e, ColumnRef) and e.col.is_computed and (not e.col.is_stored or e.col in resolve_cols)
            ])
            if len(target_col_refs) == 0:
                return result
            for ref in target_col_refs:
                assert ref.col.value_expr is not None
                result = result.substitute(ref, ref.col.value_expr)

    def is_bound_by(self, tbl: catalog.TableVersionPath) -> bool:
        """Returns True if this expr can be evaluated in the context of tbl."""
        from .column_ref import ColumnRef
        col_refs = self.subexprs(ColumnRef)
        for col_ref in col_refs:
            if not tbl.has_column(col_ref.col):
                return False
        return True

    def retarget(self, tbl: catalog.TableVersionPath) -> Expr:
        """Retarget ColumnRefs in this expr to the specific TableVersions in tbl."""
        tbl_versions = {tbl_version.id: tbl_version for tbl_version in tbl.get_tbl_versions()}
        return self._retarget(tbl_versions)

    def _retarget(self, tbl_versions: Dict[UUID, catalog.TableVersion]) -> Expr:
        from .column_ref import ColumnRef
        if isinstance(self, ColumnRef):
            target = tbl_versions[self.col.tbl.id]
            assert self.col.id in target.cols_by_id
            col = target.cols_by_id[self.col.id]
            return ColumnRef(col)
        for i in range (len(self.components)):
            self.components[i] = self.components[i]._retarget(tbl_versions)
        return self

    @classmethod
    def list_substitute(cls, expr_list: List[Expr], old: Expr, new: Expr) -> None:
        for i in range(len(expr_list)):
            expr_list[i] = expr_list[i].substitute(old, new)

    @abc.abstractmethod
    def __str__(self) -> str:
        pass

    def display_str(self, inline: bool = True) -> str:
        """
        inline: if False, use line breaks where appropriate; otherwise don't use linebreaks
        """
        return str(self)

    @classmethod
    def print_list(cls, expr_list: List[Expr]) -> str:
        if len(expr_list) == 1:
            return str(expr_list[0])
        return f'({", ".join([str(e) for e in expr_list])})'

    def subexprs(
            self, expr_class: Optional[Type[Expr]] = None, filter: Optional[Callable[[Expr], bool]] = None,
            traverse_matches: bool = True
    ) -> Generator[Expr, None, None]:
        """
        Iterate over all subexprs, including self.
        """
        assert expr_class is None or filter is None  # at most one of them
        if expr_class is not None:
            filter = lambda e: isinstance(e, expr_class)
        is_match = filter is None or filter(self)
        if not is_match or traverse_matches:
            for c in self.components:
                yield from c.subexprs(filter=filter, traverse_matches=traverse_matches)
        if is_match:
            yield self

    def contains(self, cls: Optional[Type[Expr]] = None, filter: Optional[Callable[[Expr], bool]] = None) -> bool:
        """
        Returns True if any subexpr is an instance of cls.
        """
        assert (cls is not None) != (filter is not None)  # need one of them
        if cls is not None:
            filter = lambda e: isinstance(e, cls)
        try:
            _ = next(self.subexprs(filter=filter, traverse_matches=False))
            return True
        except StopIteration:
            return False

    @classmethod
    def list_subexprs(
            cls, expr_list: List[Expr], expr_class: Optional[Type[Expr]] = None,
            filter: Optional[Callable[[Expr], bool]] = None, traverse_matches: bool = True
    ) -> Generator[Expr, None, None]:
        """Produce subexprs for all exprs in list. Can contain duplicates."""
        for e in expr_list:
            yield from e.subexprs(expr_class=expr_class, filter=filter, traverse_matches=traverse_matches)

    def tbl_ids(self) -> Set[UUID]:
        """Returns table ids referenced by this expr."""
        from .column_ref import ColumnRef
        from .rowid_ref import RowidRef
        return {ref.col.tbl.id for ref in self.subexprs(ColumnRef)} | {ref.tbl.id for ref in self.subexprs(RowidRef)}

    @classmethod
    def list_tbl_ids(cls, expr_list: List[Expr]) -> Set[UUID]:
        ids: Set[UUID] = set()
        for e in expr_list:
            ids.update(e.tbl_ids())
        return ids

    @classmethod
    def from_object(cls, o: object) -> Optional[Expr]:
        """
        Try to turn a literal object into an Expr.
        """
        if isinstance(o, Expr):
            return o
        # try to create a literal
        obj_type = ts.ColumnType.infer_literal_type(o)
        if obj_type is not None:
            from .literal import Literal
            return Literal(o, col_type=obj_type)
        if isinstance(o, dict):
            from .inline_dict import InlineDict
            return InlineDict(o)
        elif isinstance(o, list):
            from .inline_array import InlineArray
            return InlineArray(tuple(o))
        return None

    @abc.abstractmethod
    def _equals(self, other: Expr) -> bool:
        pass

    @abc.abstractmethod
    def sql_expr(self) -> Optional[sql.ClauseElement]:
        """
        If this expr can be materialized directly in SQL:
        - returns a ClauseElement
        - eval() will not be called (exception: Literal)
        Otherwise
        - returns None
        - eval() will be called
        """
        pass

    @abc.abstractmethod
    def eval(self, data_row: DataRow, row_builder: 'pixeltable.exprs.RowBuilder') -> None:
        """
        Compute the expr value for data_row and store the result in data_row[slot_idx].
        Not called if sql_expr() != None (exception: Literal).
        """
        pass

    def release(self) -> None:
        """
        Allow Expr class to tear down execution state. This is called after the last eval() call.
        """
        for c in self.components:
            c.release()

    @classmethod
    def release_list(cls, expr_list: List[Expr]) -> None:
        for e in expr_list:
            e.release()

    def serialize(self) -> str:
        return json.dumps(self.as_dict())

    def as_dict(self) -> Dict:
        """
        Turn Expr object into a dict that can be passed to json.dumps().
        Subclasses override _as_dict().
        """
        return {
            '_classname': self.__class__.__name__,
            **self._as_dict(),
        }

    @classmethod
    def as_dict_list(self, expr_list: List[Expr]) -> List[Dict]:
        return [e.as_dict() for e in expr_list]

    def _as_dict(self) -> Dict:
        if len(self.components) > 0:
            return {'components': [c.as_dict() for c in self.components]}
        return {}

    @classmethod
    def deserialize(cls, dict_str: str) -> Expr:
        return cls.from_dict(json.loads(dict_str))

    @classmethod
    def from_dict(cls, d: Dict) -> Expr:
        """
        Turn dict that was produced by calling Expr.as_dict() into an instance of the correct Expr subclass.
        """
        assert '_classname' in d
        exprs_module = importlib.import_module(cls.__module__.rsplit('.', 1)[0])
        type_class = getattr(exprs_module, d['_classname'])
        components: List[Expr] = []
        if 'components' in d:
            components = [cls.from_dict(component_dict) for component_dict in d['components']]
        return type_class._from_dict(d, components)

    @classmethod
    def from_dict_list(cls, dict_list: List[Dict]) -> List[Expr]:
        return [cls.from_dict(d) for d in dict_list]

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert False, 'not implemented'

    def astype(self, new_type: ts.ColumnType) -> 'pixeltable.exprs.TypeCast':
        from pixeltable.exprs import TypeCast
        return TypeCast(self, new_type)

    def apply(self, fn: Callable, *, col_type: Optional[ts.ColumnType] = None) -> 'pixeltable.exprs.FunctionCall':
        function = self._make_applicator_function(fn, col_type)
        # Return a `FunctionCall` obtained by passing this `Expr` to the new `function`.
        return function(self)

    def __getitem__(self, index: object) -> Expr:
        if self.col_type.is_json_type():
            from .json_path import JsonPath
            return JsonPath(self).__getitem__(index)
        if self.col_type.is_array_type():
            from .array_slice import ArraySlice
            return ArraySlice(self, index)
        raise excs.Error(f'Type {self.col_type} is not subscriptable')

    def __getattr__(self, name: str) -> Union['pixeltable.exprs.ImageMemberAccess', 'pixeltable.exprs.JsonPath']:
        """
        ex.: <img col>.rotate(60)
        """
        if self.col_type.is_image_type():
            from .image_member_access import ImageMemberAccess
            return ImageMemberAccess(name, self)
        if self.col_type.is_json_type():
            from .json_path import JsonPath
            return JsonPath(self).__getattr__(name)
        raise excs.Error(f'Member access not supported on type {self.col_type}: {name}')

    def __bool__(self) -> bool:
        raise TypeError(
            'Pixeltable expressions cannot be used in conjunction with Python boolean operators (and/or/not)')

    def __lt__(self, other: object) -> 'pixeltable.exprs.Comparison':
        return self._make_comparison(ComparisonOperator.LT, other)

    def __le__(self, other: object) -> 'pixeltable.exprs.Comparison':
        return self._make_comparison(ComparisonOperator.LE, other)

    def __eq__(self, other: object) -> 'pixeltable.exprs.Comparison':
        if other is None:
            from .is_null import IsNull
            return IsNull(self)
        return self._make_comparison(ComparisonOperator.EQ, other)

    def __ne__(self, other: object) -> 'pixeltable.exprs.Comparison':
        if other is None:
            from .compound_predicate import CompoundPredicate
            from .is_null import IsNull
            return CompoundPredicate(LogicalOperator.NOT, [IsNull(self)])
        return self._make_comparison(ComparisonOperator.NE, other)

    def __gt__(self, other: object) -> 'pixeltable.exprs.Comparison':
        return self._make_comparison(ComparisonOperator.GT, other)

    def __ge__(self, other: object) -> 'pixeltable.exprs.Comparison':
        return self._make_comparison(ComparisonOperator.GE, other)

    def _make_comparison(self, op: ComparisonOperator, other: object) -> 'pixeltable.exprs.Comparison':
        """
        other: Union[Expr, LiteralPythonTypes]
        """
        # TODO: check for compatibility
        from .comparison import Comparison
        from .literal import Literal
        if isinstance(other, Expr):
            return Comparison(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return Comparison(op, self, Literal(other))  # type: ignore[arg-type]
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def __add__(self, other: object) -> 'pixeltable.exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.ADD, other)

    def __sub__(self, other: object) -> 'pixeltable.exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.SUB, other)

    def __mul__(self, other: object) -> 'pixeltable.exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.MUL, other)

    def __truediv__(self, other: object) -> 'pixeltable.exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.DIV, other)

    def __mod__(self, other: object) -> 'pixeltable.exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.MOD, other)

    def _make_arithmetic_expr(self, op: ArithmeticOperator, other: object) -> 'pixeltable.exprs.ArithmeticExpr':
        """
        other: Union[Expr, LiteralPythonTypes]
        """
        # TODO: check for compatibility
        from .arithmetic_expr import ArithmeticExpr
        from .literal import Literal
        if isinstance(other, Expr):
            return ArithmeticExpr(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return ArithmeticExpr(op, self, Literal(other))  # type: ignore[arg-type]
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def _make_applicator_function(self, fn: Callable, col_type: Optional[ts.ColumnType]) -> 'pixeltable.func.Function':
        """
        Creates a unary pixeltable `Function` that encapsulates a python `Callable`. The result type of
        the new `Function` is given by `col_type`, and its parameter type will be `self.col_type`.

        Args:
            fn: The `Callable` to encapsulate. Must have at least one parameter, and at most one required
                parameter.
            col_type: The pixeltable result type of the new `Function`.
        """
        if col_type is not None:
            # col_type is specified explicitly
            fn_type = col_type
        elif fn in _known_applicator_types:
            # For convenience, various built-ins and other Python functions that don't
            # have type hints are hardcoded
            fn_type = _known_applicator_types[fn]
        elif 'return' in typing.get_type_hints(fn):
            # Attempt to infer the column type from the return type of the callable;
            # this will set fn_type to None if it cannot be inferred
            return_type = typing.get_type_hints(fn)['return']
            fn_type = ts.ColumnType.from_python_type(return_type)
        else:
            # No type hint
            fn_type = None

        if fn_type is None:
            raise excs.Error(
                f'Column type of `{fn.__name__}` cannot be inferred. '
                f'Use `.apply({fn.__name__}, col_type=...)` to specify.')

        # TODO(aaron-siegel) Currently we assume that `fn` has exactly one required parameter
        # and all optional parameters take their default values. Should we provide a more
        # flexible API? For example, by defining
        # expr.apply(fn, my_kw=my_arg)
        # to mean: transform each x by calling
        # fn(x, my_kw=my_arg)
        # In the current implementation, a lambda is needed in order to specify this pattern:
        # expr.apply(lambda x: fn(x, my_kw=my_arg))

        try:
            # If `fn` is not a builtin, we can do some basic validation to ensure it's
            # compatible with `apply`.
            params = inspect.signature(fn).parameters
            params_iter = iter(params.values())
            first_param = next(params_iter) if len(params) >= 1 else None
            second_param = next(params_iter) if len(params) >= 2 else None
            # Check that fn has at least one positional parameter
            if len(params) == 0 or first_param.kind in (inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD):
                raise excs.Error(
                    f'Function `{fn.__name__}` has no positional parameters.'
                )
            # Check that fn has at most one required parameter, i.e., its second parameter
            # has no default and is not a varargs
            if len(params) >= 2 and \
                    second_param.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD) and \
                    second_param.default == inspect.Parameter.empty:
                raise excs.Error(
                    f'Function `{fn.__name__}` has multiple required parameters.'
                )
        except ValueError:
            # inspect.signature(fn) will raise a `ValueError` if `fn` is a builtin; I don't
            # know of any way to get the signature of a builtin, nor to check for this in
            # advance (without the try/except pattern). For now, builtins will not be
            # validated.
            pass

        # Since `fn` might have optional parameters, we wrap it in a lambda to get a unary
        # equivalent, so that its signature is understood by `make_function`. This also
        # ensures that `decorated_fn` is never a builtin.
        # We also set the display_name explicitly, so that the `FunctionCall` gets the
        # name of `decorated_fn`, not the lambda.
        return func.make_function(
            decorated_fn=lambda x: fn(x), return_type=fn_type, param_types=[self.col_type], function_name=fn.__name__)


# A dictionary of result types of various stdlib functions that are
# commonly used in computed columns. stdlib does not have type hints, so these
# are used to infer their result types (as pixeltable types) to avoid having
# to specify them explicitly in Expr.apply().
# This is purely for convenience and does not impact the supported functionality
# (it's always possible to specify a result type explicitly for a function
# that does not have type hints and is not present in this dict).
_known_applicator_types: dict[Callable, ts.ColumnType] = {
    str: ts.StringType(),
    json.dumps: ts.StringType(),
    json.loads: ts.JsonType(),
}



from typing import Any, Callable

import PIL.Image
import numpy as np

import pixeltable as pxt
import pixeltable.env as env
import pixeltable.type_system as ts
from pixeltable.func import Batch


@pxt.udf(batch_size=32, return_type=ts.ArrayType((None,), dtype=ts.FloatType()))
def sentence_transformer(sentences: Batch[str], *, model_id: str, normalize_embeddings: bool = False) -> Batch[np.ndarray]:
    env.Env.get().require_package('sentence_transformers')
    from sentence_transformers import SentenceTransformer

    model = _lookup_model(model_id, SentenceTransformer)

    array = model.encode(sentences, normalize_embeddings=normalize_embeddings)
    return [array[i] for i in range(array.shape[0])]


@pxt.udf
def sentence_transformer_list(sentences: list, *, model_id: str, normalize_embeddings: bool = False) -> list:
    env.Env.get().require_package('sentence_transformers')
    from sentence_transformers import SentenceTransformer

    model = _lookup_model(model_id, SentenceTransformer)

    array = model.encode(sentences, normalize_embeddings=normalize_embeddings)
    return [array[i].tolist() for i in range(array.shape[0])]


@pxt.udf(batch_size=32)
def cross_encoder(sentences1: Batch[str], sentences2: Batch[str], *, model_id: str) -> Batch[float]:
    env.Env.get().require_package('sentence_transformers')
    from sentence_transformers import CrossEncoder

    model = _lookup_model(model_id, CrossEncoder)

    array = model.predict([[s1, s2] for s1, s2 in zip(sentences1, sentences2)], convert_to_numpy=True)
    return array.tolist()


@pxt.udf
def cross_encoder_list(sentence1: str, sentences2: list, *, model_id: str) -> list:
    env.Env.get().require_package('sentence_transformers')
    from sentence_transformers import CrossEncoder

    model = _lookup_model(model_id, CrossEncoder)

    array = model.predict([[sentence1, s2] for s2 in sentences2], convert_to_numpy=True)
    return array.tolist()


@pxt.udf(batch_size=32, return_type=ts.ArrayType((None,), dtype=ts.FloatType(), nullable=False))
def clip_text(text: Batch[str], *, model_id: str) -> Batch[np.ndarray]:
    env.Env.get().require_package('transformers')
    from transformers import CLIPModel, CLIPProcessor

    model = _lookup_model(model_id, CLIPModel.from_pretrained)
    processor = _lookup_processor(model_id, CLIPProcessor.from_pretrained)

    inputs = processor(text=text, return_tensors='pt', padding=True, truncation=True)
    embeddings = model.get_text_features(**inputs).detach().numpy()
    return [embeddings[i] for i in range(embeddings.shape[0])]


@pxt.udf(batch_size=32, return_type=ts.ArrayType((None,), dtype=ts.FloatType(), nullable=False))
def clip_image(image: Batch[PIL.Image.Image], *, model_id: str) -> Batch[np.ndarray]:
    env.Env.get().require_package('transformers')
    from transformers import CLIPModel, CLIPProcessor

    model = _lookup_model(model_id, CLIPModel.from_pretrained)
    processor = _lookup_processor(model_id, CLIPProcessor.from_pretrained)

    inputs = processor(images=image, return_tensors='pt', padding=True)
    embeddings = model.get_image_features(**inputs).detach().numpy()
    return [embeddings[i] for i in range(embeddings.shape[0])]


@pxt.udf(batch_size=32)
def detr_for_object_detection(image: Batch[PIL.Image.Image], *, model_id: str, threshold: float = 0.5) -> Batch[dict]:
    env.Env.get().require_package('transformers')
    from transformers import DetrImageProcessor, DetrForObjectDetection

    model = _lookup_model(model_id, lambda x: DetrForObjectDetection.from_pretrained(x, revision='no_timm'))
    processor = _lookup_processor(model_id, lambda x: DetrImageProcessor.from_pretrained(x, revision='no_timm'))

    inputs = processor(images=image, return_tensors='pt')
    outputs = model(**inputs)

    results = processor.post_process_object_detection(outputs, threshold=threshold)
    return [
        {
            'scores': [score.item() for score in result['scores']],
            'labels': [label.item() for label in result['labels']],
            'label_text': [model.config.id2label[label.item()] for label in result['labels']],
            'boxes': [box.tolist() for box in result['boxes']]
        }
        for result in results
    ]


def _lookup_model(model_id: str, create: Callable) -> Any:
    key = (model_id, create)  # For safety, include the `create` callable in the cache key
    if key not in _model_cache:
        _model_cache[key] = create(model_id)
    return _model_cache[key]


def _lookup_processor(model_id: str, create: Callable) -> Any:
    key = (model_id, create)  # For safety, include the `create` callable in the cache key
    if key not in _processor_cache:
        _processor_cache[key] = create(model_id)
    return _processor_cache[key]


_model_cache = {}
_processor_cache = {}



from typing import Optional, List, Tuple

import PIL
import pytest

import pixeltable as pxt
from pixeltable import catalog
from pixeltable import exceptions as excs
from pixeltable.iterators import FrameIterator
from pixeltable.tests.utils import get_video_files
from pixeltable.type_system import VideoType, ImageType
from pixeltable.utils.media_store import MediaStore


class TestVideo:
    def create_tbls(
            self, cl: pxt.Client, base_name: str = 'video_tbl', view_name: str = 'frame_view'
    ) -> Tuple[catalog.InsertableTable, catalog.Table]:
        cl.drop_table(view_name, ignore_errors=True)
        cl.drop_table(base_name, ignore_errors=True)
        base_t = cl.create_table(base_name, {'video': VideoType()})
        args = {'video': base_t.video, 'fps': 1}
        view_t = cl.create_view(view_name, base_t, iterator_class=FrameIterator, iterator_args=args)
        return base_t, view_t

    def create_and_insert(
            self, cl: pxt.Client, stored: Optional[bool], paths: List[str]
    ) -> Tuple[catalog.InsertableTable, catalog.Table]:
        base_t, view_t = self.create_tbls(cl)

        view_t.add_column(transform=view_t.frame.rotate(90), stored=stored)
        base_t.insert({'video': p} for p in paths)
        total_num_rows = view_t.count()
        result = view_t[view_t.frame_idx >= 5][view_t.frame_idx, view_t.frame, view_t.transform].show(0)
        assert len(result) == total_num_rows - len(paths) * 5
        result = view_t[view_t.frame_idx, view_t.frame, view_t.transform].show(3)
        assert len(result) == 3
        result = view_t[view_t.frame_idx, view_t.frame, view_t.transform].show(0)
        assert len(result) == total_num_rows
        return base_t, view_t

    def test_basic(self, test_client: pxt.Client) -> None:
        video_filepaths = get_video_files()
        cl = test_client

        # default case: computed images are not stored
        _, view = self.create_and_insert(cl, None, video_filepaths)
        assert MediaStore.count(view.get_id()) == 0

        # computed images are explicitly not stored
        _, view = self.create_and_insert(cl, False, video_filepaths)
        assert MediaStore.count(view.get_id()) == 0

        # computed images are stored
        tbl, view = self.create_and_insert(cl, True, video_filepaths)
        assert MediaStore.count(view.get_id()) == view.count()

        # revert() also removes computed images
        tbl.insert({'video': p} for p in video_filepaths)
        tbl.revert()
        assert MediaStore.count(view.get_id()) == view.count()

    def test_query(self, test_client: pxt.client) -> None:
        video_filepaths = get_video_files()
        cl = test_client
        base_t, view_t = self.create_tbls(cl)
        # also include an external file, to make sure that prefetching works
        url = 's3://multimedia-commons/data/videos/mp4/ffe/ff3/ffeff3c6bf57504e7a6cecaff6aefbc9.mp4'
        video_filepaths.append(url)
        status = base_t.insert({'video': p} for p in video_filepaths)
        assert status.num_excs == 0
        # make sure that we can get the frames back
        res = view_t.select(view_t.frame).collect().to_pandas()
        assert res['frame'].notnull().all()
        # make sure we can select a specific video
        all_rows = view_t.select(url=view_t.video.fileurl).collect().to_pandas()
        res = view_t.where(view_t.video == url).collect()
        assert len(res) == len(all_rows[all_rows.url == url])

    def test_fps(self, test_client: pxt.client) -> None:
        cl = test_client
        path = get_video_files()[0]
        videos = cl.create_table('videos', {'video': VideoType()})
        frames_1_0 = cl.create_view(
            'frames_1_0', videos, iterator_class=FrameIterator, iterator_args={'video': videos.video, 'fps': 1})
        frames_0_5 = cl.create_view(
            'frames_0_5', videos, iterator_class=FrameIterator, iterator_args={'video': videos.video, 'fps': 1/2})
        frames_0_33 = cl.create_view(
            'frames_0_33', videos, iterator_class=FrameIterator, iterator_args={'video': videos.video, 'fps': 1/3})
        videos.insert(video=path)
        assert frames_0_5.count() == frames_1_0.count() // 2 or frames_0_5.count() == frames_1_0.count() // 2 + 1
        assert frames_0_33.count() == frames_1_0.count() // 3 or frames_0_33.count() == frames_1_0.count() // 3 + 1

    def test_computed_cols(self, test_client: pxt.client) -> None:
        video_filepaths = get_video_files()
        cl = test_client
        base_t, view_t = self.create_tbls(cl)
        # c2 and c4 depend directly on c1, c3 depends on it indirectly
        view_t.add_column(c1=view_t.frame.resize([224, 224]))
        view_t.add_column(c2=view_t.c1.rotate(10))
        view_t.add_column(c3=view_t.c2.rotate(20))
        view_t.add_column(c4=view_t.c1.rotate(30))
        for name in ['c1', 'c2', 'c3', 'c4']:
            assert not view_t.tbl_version_path.tbl_version.cols_by_name[name].is_stored
        base_t.insert({'video': p} for p in video_filepaths)
        _ = view_t[view_t.c1, view_t.c2, view_t.c3, view_t.c4].show(0)

    def test_make_video(self, test_client: pxt.Client) -> None:
        video_filepaths = get_video_files()
        cl = test_client
        base_t, view_t = self.create_tbls(cl)
        base_t.insert({'video': p} for p in video_filepaths)
        # reference to the frame col requires ordering by base, pos
        from pixeltable.functions import make_video
        _ = view_t.select(make_video(view_t.pos, view_t.frame)).group_by(base_t).show()
        # the same without frame col
        view_t.add_column(transformed=view_t.frame.rotate(30), stored=True)
        _ = view_t.select(make_video(view_t.pos, view_t.transformed)).group_by(base_t).show()

        with pytest.raises(excs.Error):
            # make_video() doesn't allow windows
            _ = view_t.select(make_video(view_t.pos, view_t.frame, group_by=base_t)).show()
        with pytest.raises(excs.Error):
            # make_video() requires ordering
            _ = view_t.select(make_video(view_t.frame, order_by=view_t.pos)).show()
        with pytest.raises(excs.Error):
            # incompatible ordering requirements
            _ = view_t.select(
                make_video(view_t.pos, view_t.frame),
                make_video(view_t.pos - 1, view_t.transformed)).group_by(base_t).show()

        # window function that simply passes through the frame
        @pxt.uda(
            update_types=[ImageType()], value_type=ImageType(), name='agg_fn',
            requires_order_by=True, allows_std_agg=False, allows_window=True)
        class WindowAgg:
            def __init__(self):
                self.img = None
            def update(self, frame: PIL.Image.Image) -> None:
                self.img = frame
            def value(self) -> PIL.Image.Image:
                return self.img

        # make sure it works
        _ = view_t.select(agg_fn(view_t.pos, view_t.frame, group_by=base_t)).show()
        status = view_t.add_column(agg=agg_fn(view_t.pos, view_t.frame, group_by=base_t))
        assert status.num_excs == 0
        _ = view_t.select(make_video(view_t.pos, view_t.agg)).group_by(base_t).show()

        # image cols computed with a window function currently need to be stored
        with pytest.raises(excs.Error):
            view_t.add_column(agg2=agg_fn(view_t.pos, view_t.frame, group_by=base_t), stored=False)

        # reload from store
        cl = pxt.Client(reload=True)
        base_t, view_t = cl.get_table(base_t.get_name()), cl.get_table(view_t.get_name())
        _ = view_t.select(agg_fn(view_t.pos, view_t.frame, group_by=base_t)).show()



from typing import Tuple, Optional, List, Set, Any, Dict
from uuid import UUID

import sqlalchemy as sql

import pixeltable.exec as exec
import pixeltable.func as func
from pixeltable import catalog
from pixeltable import exceptions as excs
from pixeltable import exprs


def _is_agg_fn_call(e: exprs.Expr) -> bool:
    return isinstance(e, exprs.FunctionCall) and e.is_agg_fn_call and not e.is_window_fn_call

def _get_combined_ordering(
        o1: List[Tuple[exprs.Expr, bool]], o2: List[Tuple[exprs.Expr, bool]]
) -> List[Tuple[exprs.Expr, bool]]:
    """Returns an ordering that's compatible with both o1 and o2, or an empty list if no such ordering exists"""
    result: List[Tuple[exprs.Expr, bool]] = []
    # determine combined ordering
    for (e1, asc1), (e2, asc2) in zip(o1, o2):
        if e1.id != e2.id:
            return []
        if asc1 is not None and asc2 is not None and asc1 != asc2:
            return []
        asc = asc1 if asc1 is not None else asc2
        result.append((e1, asc))

    # add remaining ordering of the longer list
    prefix_len = min(len(o1), len(o2))
    if len(o1) > prefix_len:
        result.extend(o1[prefix_len:])
    elif len(o2) > prefix_len:
        result.extend(o2[prefix_len:])
    return result

class Analyzer:
    """Class to perform semantic analysis of a query and to store the analysis state"""

    def __init__(
            self, tbl: catalog.TableVersionPath, select_list: List[exprs.Expr],
            where_clause: Optional[exprs.Predicate] = None, group_by_clause: Optional[List[exprs.Expr]] = None,
            order_by_clause: Optional[List[Tuple[exprs.Expr, bool]]] = None):
        if group_by_clause is None:
            group_by_clause = []
        if order_by_clause is None:
            order_by_clause = []
        self.tbl = tbl

        # remove references to unstored computed cols
        self.select_list = [e.resolve_computed_cols() for e in select_list]
        if where_clause is not None:
            where_clause = where_clause.resolve_computed_cols()
        self.group_by_clause = [e.resolve_computed_cols() for e in group_by_clause]
        self.order_by_clause = [(e.resolve_computed_cols(), asc) for e, asc in order_by_clause]

        # Where clause of the Select stmt of the SQL scan
        self.sql_where_clause: Optional[exprs.Expr] = None
        # filter predicate applied to output rows of the SQL scan
        self.filter: Optional[exprs.Predicate] = None
        # not executable
        self.similarity_clause: Optional[exprs.ImageSimilarityPredicate] = None
        if where_clause is not None:
            where_clause_conjuncts, self.filter = where_clause.split_conjuncts(lambda e: e.sql_expr() is not None)
            self.sql_where_clause = exprs.CompoundPredicate.make_conjunction(where_clause_conjuncts)
            if self.filter is not None:
                similarity_clauses, self.filter = self.filter.split_conjuncts(
                    lambda e: isinstance(e, exprs.ImageSimilarityPredicate))
                if len(similarity_clauses) > 1:
                    raise excs.Error(f'More than one nearest() not supported')
                if len(similarity_clauses) == 1:
                    if len(self.order_by_clause) > 0:
                        raise excs.Error((
                            f'nearest() returns results in order of proximity and cannot be used in conjunction with '
                            f'order_by()'))
                    self.similarity_clause = similarity_clauses[0]
                    img_col = self.similarity_clause.img_col_ref.col
                    if not img_col.is_indexed:
                        raise excs.Error(f'nearest() not available for unindexed column {img_col.name}')

        # all exprs that are evaluated in Python; not executable
        self.all_exprs = self.select_list.copy()
        self.all_exprs.extend(self.group_by_clause)
        self.all_exprs.extend([e for e, _ in self.order_by_clause])
        if self.filter is not None:
            self.all_exprs.append(self.filter)
        self.sql_exprs = list(exprs.Expr.list_subexprs(
            self.all_exprs, filter=lambda e: e.sql_expr() is not None, traverse_matches=False))

        # sql_exprs: exprs that can be expressed via SQL and are retrieved directly from the store
        # (we don't want to materialize literals via SQL, so we remove them here)
        self.sql_exprs = [e for e in self.sql_exprs if not isinstance(e, exprs.Literal)]

        self.agg_fn_calls: List[exprs.FunctionCall] = []
        self.agg_order_by: List[exprs.Expr] = []
        self._analyze_agg()

    def _analyze_agg(self) -> None:
        """Check semantic correctness of aggregation and fill in agg-specific fields of Analyzer"""
        self.agg_fn_calls = [e for e in self.all_exprs if _is_agg_fn_call(e)]
        if len(self.agg_fn_calls) == 0:
            # nothing to do
            return

        # check that select list only contains aggregate output
        grouping_expr_ids = {e.id for e in self.group_by_clause}
        is_agg_output = [self._determine_agg_status(e, grouping_expr_ids)[0] for e in self.select_list]
        if is_agg_output.count(False) > 0:
            raise excs.Error(
                f'Invalid non-aggregate expression in aggregate query: {self.select_list[is_agg_output.index(False)]}')

        # check that filter doesn't contain aggregates
        if self.filter is not None:
            agg_fn_calls = [e for e in self.filter.subexprs(filter=lambda e: _is_agg_fn_call(e))]
            if len(agg_fn_calls) > 0:
                raise excs.Error(f'Filter cannot contain aggregate functions: {self.filter}')

        # check that grouping exprs don't contain aggregates and can be expressed as SQL (we perform sort-based
        # aggregation and rely on the SqlScanNode returning data in the correct order)
        for e in self.group_by_clause:
            if e.sql_expr() is None:
                raise excs.Error(f'Invalid grouping expression, needs to be expressible in SQL: {e}')
            if e.contains(filter=lambda e: _is_agg_fn_call(e)):
                raise excs.Error(f'Grouping expression contains aggregate function: {e}')

        # check that agg fn calls don't have contradicting ordering requirements
        order_by: List[exprs.Exprs] = []
        order_by_origin: Optional[exprs.Expr] = None  # the expr that determines the ordering
        for agg_fn_call in self.agg_fn_calls:
            fn_call_order_by = agg_fn_call.get_agg_order_by()
            if len(fn_call_order_by) == 0:
                continue
            if len(order_by) == 0:
                order_by = fn_call_order_by
                order_by_origin = agg_fn_call
            else:
                combined = _get_combined_ordering(
                    [(e, True) for e in order_by], [(e, True) for e in fn_call_order_by])
                if len(combined) == 0:
                    raise excs.Error((
                        f"Incompatible ordering requirements between expressions '{order_by_origin}' and "
                        f"'{agg_fn_call}':\n"
                        f"{exprs.Expr.print_list(order_by)} vs {exprs.Expr.print_list(fn_call_order_by)}"
                    ))
        self.agg_order_by = order_by

    def _determine_agg_status(self, e: exprs.Expr, grouping_expr_ids: Set[int]) -> Tuple[bool, bool]:
        """Determine whether expr is the input to or output of an aggregate function.
        Returns:
            (<is output>, <is input>)
        """
        if e.id in grouping_expr_ids:
            return True, True
        elif _is_agg_fn_call(e):
            for c in e.components:
                _, is_input = self._determine_agg_status(c, grouping_expr_ids)
                if not is_input:
                    raise excs.Error(f'Invalid nested aggregates: {e}')
            return True, False
        elif isinstance(e, exprs.Literal):
            return True, True
        elif isinstance(e, exprs.ColumnRef) or isinstance(e, exprs.RowidRef):
            # we already know that this isn't a grouping expr
            return False, True
        else:
            # an expression such as <grouping expr 1> + <grouping expr 2> can both be the output and input of agg
            assert len(e.components) > 0
            component_is_output, component_is_input = zip(
                *[self._determine_agg_status(c, grouping_expr_ids) for c in e.components])
            is_output = component_is_output.count(True) == len(e.components)
            is_input = component_is_input.count(True) == len(e.components)
            if not is_output and not is_input:
                raise excs.Error(f'Invalid expression, mixes aggregate with non-aggregate: {e}')
            return is_output, is_input


    def finalize(self, row_builder: exprs.RowBuilder) -> None:
        """Make all exprs executable
        TODO: add EvalCtx for each expr list?
        """
        # maintain original composition of select list
        row_builder.substitute_exprs(self.select_list, remove_duplicates=False)
        row_builder.substitute_exprs(self.group_by_clause)
        order_by_exprs = [e for e, _ in self.order_by_clause]
        row_builder.substitute_exprs(order_by_exprs)
        self.order_by_clause = [(e, asc) for e, (_, asc) in zip(order_by_exprs, self.order_by_clause)]
        row_builder.substitute_exprs(self.all_exprs)
        row_builder.substitute_exprs(self.sql_exprs)
        if self.filter is not None:
            self.filter = row_builder.unique_exprs[self.filter]
        row_builder.substitute_exprs(self.agg_fn_calls)
        row_builder.substitute_exprs(self.agg_order_by)


class Planner:
    # TODO: create an exec.CountNode and change this to create_count_plan()
    @classmethod
    def create_count_stmt(
            cls, tbl: catalog.TableVersionPath, where_clause: Optional[exprs.Predicate] = None
    ) -> sql.Select:
        stmt = sql.select(sql.func.count('*'))
        refd_tbl_ids: Set[UUID] = set()
        if where_clause is not None:
            analyzer = cls.analyze(tbl, where_clause)
            if analyzer.similarity_clause is not None:
                raise excs.Error('nearest() cannot be used with count()')
            if analyzer.filter is not None:
                raise excs.Error(f'Filter {analyzer.filter} not expressible in SQL')
            clause_element = analyzer.sql_where_clause.sql_expr()
            assert clause_element is not None
            stmt = stmt.where(clause_element)
            refd_tbl_ids = where_clause.tbl_ids()
        stmt = exec.SqlScanNode.create_from_clause(tbl, stmt, refd_tbl_ids)
        return stmt

    @classmethod
    def create_insert_plan(
            cls, tbl: catalog.TableVersion, rows: List[Dict[str, Any]], ignore_errors: bool
    ) -> exec.ExecNode:
        """Creates a plan for TableVersion.insert()"""
        assert not tbl.is_view()
        # things we need to materialize:
        # 1. stored_cols: all cols we need to store, incl computed cols (and indices)
        stored_cols = [c for c in tbl.cols if c.is_stored]
        assert len(stored_cols) > 0
        # 2. values to insert into indices
        indexed_cols = [c for c in tbl.cols if c.is_indexed]
        index_info: List[Tuple[catalog.Column, func.Function]] = []
        if len(indexed_cols) > 0:
            from pixeltable.functions.nos.image_embedding import openai_clip
            index_info = [(c, openai_clip) for c in tbl.cols if c.is_indexed]

        row_builder = exprs.RowBuilder([], stored_cols, index_info, [])

        # create InMemoryDataNode for 'rows'
        stored_col_info = row_builder.output_slot_idxs()
        stored_img_col_info = [info for info in stored_col_info if info.col.col_type.is_image_type()]
        input_col_info = [info for info in stored_col_info if not info.col.is_computed]
        plan = exec.InMemoryDataNode(tbl, rows, row_builder, tbl.next_rowid)

        media_input_cols = [info for info in input_col_info if info.col.col_type.is_media_type()]

        # prefetch external files for all input column refs for validation
        plan = exec.CachePrefetchNode(tbl.id, media_input_cols, plan)
        plan = exec.MediaValidationNode(row_builder, media_input_cols, input=plan)

        computed_exprs = row_builder.default_eval_ctx.target_exprs
        if len(computed_exprs) > 0:
            # add an ExprEvalNode when there are exprs to compute
            plan = exec.ExprEvalNode(row_builder, computed_exprs, [], input=plan)

        plan.set_stored_img_cols(stored_img_col_info)
        plan.set_ctx(
            exec.ExecContext(
                row_builder, batch_size=0, show_pbar=True, num_computed_exprs=len(computed_exprs),
                ignore_errors=ignore_errors))
        return plan

    @classmethod
    def create_update_plan(
            cls, tbl: catalog.TableVersionPath,
            update_targets: List[Tuple[catalog.Column, exprs.Expr]],
            recompute_targets: List[catalog.Column],
            where_clause: Optional[exprs.Predicate], cascade: bool
    ) -> Tuple[exec.ExecNode, List[str], List[catalog.Column]]:
        """Creates a plan to materialize updated rows.
        The plan:
        - retrieves rows that are visible at the current version of the table
        - materializes all stored columns and the update targets
        - if cascade is True, recomputes all computed columns that transitively depend on the updated columns
          and copies the values of all other stored columns
        - if cascade is False, copies all columns that aren't update targets from the original rows
        Returns:
            - root node of the plan
            - list of qualified column names that are getting updated
            - list of columns that are being recomputed
        """
        # retrieve all stored cols and all target exprs
        assert isinstance(tbl, catalog.TableVersionPath)
        target = tbl.tbl_version  # the one we need to update
        updated_cols = [col for col, _ in update_targets]
        if len(recompute_targets) > 0:
            recomputed_cols = recompute_targets.copy()
        else:
            recomputed_cols = target.get_dependent_columns(updated_cols) if cascade else {}
            # we only need to recompute stored columns (unstored ones are substituted away)
            recomputed_cols = {c for c in recomputed_cols if c.is_stored}
        recomputed_base_cols = {col for col in recomputed_cols if col.tbl == target}
        copied_cols = [
            col for col in target.cols if col.is_stored and not col in updated_cols and not col in recomputed_base_cols
        ]
        select_list = [exprs.ColumnRef(col) for col in copied_cols]
        select_list.extend([expr for _, expr in update_targets])

        recomputed_exprs = \
            [c.value_expr.copy().resolve_computed_cols(resolve_cols=recomputed_base_cols) for c in recomputed_base_cols]
        # recomputed cols reference the new values of the updated cols
        for col, e in update_targets:
            exprs.Expr.list_substitute(recomputed_exprs, exprs.ColumnRef(col), e)
        select_list.extend(recomputed_exprs)

        # we need to retrieve the PK columns of the existing rows
        plan = cls.create_query_plan(tbl, select_list, where_clause=where_clause, with_pk=True, ignore_errors=True)
        all_base_cols = copied_cols + updated_cols + list(recomputed_base_cols)  # same order as select_list
        # update row builder with column information
        [plan.row_builder.add_table_column(col, select_list[i].slot_idx) for i, col in enumerate(all_base_cols)]
        return plan, [f'{c.tbl.name}.{c.name}' for c in updated_cols + list(recomputed_cols)], list(recomputed_cols)

    @classmethod
    def create_view_update_plan(
            cls, view: catalog.TableVersionPath, recompute_targets: List[catalog.Column]
    ) -> exec.ExecNode:
        """Creates a plan to materialize updated rows for a view, given that the base table has been updated.
        The plan:
        - retrieves rows that are visible at the current version of the table and satisfy the view predicate
        - materializes all stored columns and the update targets
        - if cascade is True, recomputes all computed columns that transitively depend on the updated columns
          and copies the values of all other stored columns
        - if cascade is False, copies all columns that aren't update targets from the original rows

        TODO: unify with create_view_load_plan()

        Returns:
            - root node of the plan
            - list of qualified column names that are getting updated
            - list of columns that are being recomputed
        """
        assert isinstance(view, catalog.TableVersionPath)
        assert view.is_view()
        target = view.tbl_version  # the one we need to update
        # retrieve all stored cols and all target exprs
        recomputed_cols = set(recompute_targets.copy())
        copied_cols = [col for col in target.cols if col.is_stored and not col in recomputed_cols]
        select_list = [exprs.ColumnRef(col) for col in copied_cols]
        # resolve recomputed exprs to stored columns in the base
        recomputed_exprs = \
            [c.value_expr.copy().resolve_computed_cols(resolve_cols=recomputed_cols) for c in recomputed_cols]
        select_list.extend(recomputed_exprs)

        # we need to retrieve the PK columns of the existing rows
        plan = cls.create_query_plan(
            view, select_list, where_clause=target.predicate, with_pk=True, ignore_errors=True,
            exact_version_only=view.get_bases())
        [
            plan.row_builder.add_table_column(col, select_list[i].slot_idx)
            for i, col in enumerate(copied_cols + list(recomputed_cols))  # same order as select_list
        ]
        # TODO: avoid duplication with view_load_plan() logic (where does this belong?)
        stored_img_col_info = \
            [info for info in plan.row_builder.output_slot_idxs() if info.col.col_type.is_image_type()]
        plan.set_stored_img_cols(stored_img_col_info)
        return plan

    @classmethod
    def create_view_load_plan(
            cls, view: catalog.TableVersionPath, propagates_insert: bool = False
    ) -> Tuple[exec.ExecNode, int]:
        """Creates a query plan for populating a view.

        Args:
            view: the view to populate
            propagates_insert: if True, we're propagating a base update to this view

        Returns:
            - root node of the plan
            - number of materialized values per row
        """
        assert isinstance(view, catalog.TableVersionPath)
        assert view.is_view()
        # things we need to materialize as DataRows:
        # 1. stored computed cols
        # - iterator columns are effectively computed, just not with a value_expr
        # - we can ignore stored non-computed columns because they have a default value that is supplied directly by
        #   the store
        target = view.tbl_version  # the one we need to populate
        stored_cols = [c for c in target.cols if c.is_stored and (c.is_computed or target.is_iterator_column(c))]
        # 2. index values
        indexed_cols = [c for c in target.cols if c.is_indexed]
        index_info: List[Tuple[catalog.Column, func.Function]] = []
        if len(indexed_cols) > 0:
            from pixeltable.functions.nos.image_embedding import openai_clip
            index_info = [(c, openai_clip) for c in target.cols if c.is_indexed]
        # 3. for component views: iterator args
        iterator_args = [target.iterator_args] if target.iterator_args is not None else []

        row_builder = exprs.RowBuilder(iterator_args, stored_cols, index_info, [])

        # execution plan:
        # 1. materialize exprs computed from the base that are needed for stored view columns
        # 2. if it's an iterator view, expand the base rows into component rows
        # 3. materialize stored view columns that haven't been produced by step 1
        base_output_exprs = [e for e in row_builder.default_eval_ctx.exprs if e.is_bound_by(view.base)]
        view_output_exprs = [
            e for e in row_builder.default_eval_ctx.target_exprs
            if e.is_bound_by(view) and not e.is_bound_by(view.base)
        ]
        # if we're propagating an insert, we only want to see those base rows that were created for the current version
        base_analyzer = Analyzer(view, base_output_exprs, where_clause=target.predicate)
        plan = cls._create_query_plan(
            view.base, row_builder=row_builder, analyzer=base_analyzer, with_pk=True,
            exact_version_only=view.get_bases() if propagates_insert else [])
        exec_ctx = plan.ctx
        if target.is_component_view():
            plan = exec.ComponentIterationNode(target, plan)
        if len(view_output_exprs) > 0:
            plan = exec.ExprEvalNode(
                row_builder, output_exprs=view_output_exprs, input_exprs=base_output_exprs,input=plan)

        stored_img_col_info = [info for info in row_builder.output_slot_idxs() if info.col.col_type.is_image_type()]
        plan.set_stored_img_cols(stored_img_col_info)
        exec_ctx.ignore_errors = True
        plan.set_ctx(exec_ctx)
        return plan, len(row_builder.default_eval_ctx.target_exprs)

    @classmethod
    def _determine_ordering(cls, analyzer: Analyzer) -> List[Tuple[exprs.Expr, bool]]:
        """Returns the exprs for the ORDER BY clause of the SqlScanNode"""
        order_by_items: List[Tuple[exprs.Expr, Optional[bool]]] = []
        order_by_origin: Optional[exprs.Expr] = None  # the expr that determines the ordering


        # window functions require ordering by the group_by/order_by clauses
        window_fn_calls = [
            e for e in analyzer.all_exprs if isinstance(e, exprs.FunctionCall) and e.is_window_fn_call
        ]
        if len(window_fn_calls) > 0:
            for fn_call in window_fn_calls:
                gb, ob = fn_call.get_window_sort_exprs()
                # for now, the ordering is implicitly ascending
                fn_call_ordering = [(e, None) for e in gb] + [(e, True) for e in ob]
                if len(order_by_items) == 0:
                    order_by_items = fn_call_ordering
                    order_by_origin = fn_call
                else:
                    # check for compatibility
                    other_order_by_clauses = fn_call_ordering
                    combined = _get_combined_ordering(order_by_items, other_order_by_clauses)
                    if len(combined) == 0:
                        raise excs.Error((
                            f"Incompatible ordering requirements between expressions '{order_by_origin}' and "
                            f"'{fn_call}':\n"
                            f"{exprs.Expr.print_list(order_by_items)} vs {exprs.Expr.print_list(other_order_by_clauses)}"
                        ))
                    order_by_items = combined

        if len(analyzer.group_by_clause) > 0:
            agg_ordering = [(e, None) for e in analyzer.group_by_clause] + [(e, True) for e in analyzer.agg_order_by]
            if len(order_by_items) > 0:
                # check for compatibility
                combined = _get_combined_ordering(order_by_items, agg_ordering)
                if len(combined) == 0:
                    raise excs.Error((
                        f"Incompatible ordering requirements between expressions '{order_by_origin}' and "
                        f"grouping expressions:\n"
                        f"{exprs.Expr.print_list([e for e, _ in order_by_items])} vs "
                        f"{exprs.Expr.print_list([e for e, _ in agg_ordering])}"
                    ))
                order_by_items = combined
            else:
                order_by_items = agg_ordering

        if len(analyzer.order_by_clause) > 0:
            if len(order_by_items) > 0:
                # check for compatibility
                combined = _get_combined_ordering(order_by_items, analyzer.order_by_clause)
                if len(combined) == 0:
                    raise excs.Error((
                        f"Incompatible ordering requirements between expressions '{order_by_origin}' and "
                        f"order-by expressions:\n"
                        f"{exprs.Expr.print_list([e for e, _ in order_by_items])} vs "
                        f"{exprs.Expr.print_list([e for e, _ in analyzer.order_by_clause])}"
                    ))
                order_by_items = combined
            else:
                order_by_items = analyzer.order_by_clause

        # TODO: can this be unified with the same logic in RowBuilder
        def refs_unstored_iter_col(e: exprs.Expr) -> bool:
            if not isinstance(e, exprs.ColumnRef):
                return False
            tbl = e.col.tbl
            return tbl.is_component_view() and tbl.is_iterator_column(e.col) and not e.col.is_stored
        unstored_iter_col_refs = list(exprs.Expr.list_subexprs(analyzer.all_exprs, filter=refs_unstored_iter_col))
        if len(unstored_iter_col_refs) > 0 and len(order_by_items) == 0:
            # we don't already have a user-requested ordering and we access unstored iterator columns:
            # order by the primary key of the component view, which minimizes the number of iterator instantiations
            component_views = {e.col.tbl for e in unstored_iter_col_refs}
            # TODO: generalize this to multi-level iteration
            assert len(component_views) == 1
            component_view = list(component_views)[0]
            order_by_items = [
                (exprs.RowidRef(component_view, idx), None)
                for idx in range(len(component_view.store_tbl.rowid_columns()))
            ]
            order_by_origin = unstored_iter_col_refs[0]

        for e in [e for e, _ in order_by_items]:
            if e.sql_expr() is None:
                raise excs.Error(f'order_by element cannot be expressed in SQL: {e}')
        # we do ascending ordering by default, if not specified otherwise
        order_by_items = [(e, True) if asc is None else (e, asc) for e, asc in order_by_items]
        return order_by_items

    @classmethod
    def _is_contained_in(cls, l1: List[exprs.Expr], l2: List[exprs.Expr]) -> bool:
        """Returns True if l1 is contained in l2"""
        s1, s2 = set([e.id for e in l1]), set([e.id for e in l2])
        return s1 <= s2

    @classmethod
    def _insert_prefetch_node(
            cls, tbl_id: UUID, output_exprs: List[exprs.Expr], row_builder: exprs.RowBuilder, input: exec.ExecNode
    ) -> exec.ExecNode:
        """Returns a CachePrefetchNode into the plan if needed, otherwise returns input"""
        # we prefetch external files for all media ColumnRefs, even those that aren't part of the dependencies
        # of output_exprs: if unstored iterator columns are present, we might need to materialize ColumnRefs that
        # aren't explicitly captured as dependencies
        media_col_refs = [
            e for e in list(row_builder.unique_exprs) if isinstance(e, exprs.ColumnRef) and e.col_type.is_media_type()
        ]
        if len(media_col_refs) == 0:
            return input
        # we need to prefetch external files for media column types
        file_col_info = [exprs.ColumnSlotIdx(e.col, e.slot_idx) for e in media_col_refs]
        prefetch_node = exec.CachePrefetchNode(tbl_id, file_col_info, input)
        return prefetch_node

    @classmethod
    def create_query_plan(
            cls, tbl: catalog.TableVersionPath, select_list: Optional[List[exprs.Expr]] = None,
            where_clause: Optional[exprs.Predicate] = None, group_by_clause: Optional[List[exprs.Expr]] = None,
            order_by_clause: Optional[List[Tuple[exprs.Expr, bool]]] = None, limit: Optional[int] = None,
            with_pk: bool = False, ignore_errors: bool = False, exact_version_only: Optional[List[catalog.TableVersion]] = None
    ) -> exec.ExecNode:
        """Return plan for executing a query.
        Updates 'select_list' in place to make it executable.
        TODO: make exact_version_only a flag and use the versions from tbl
        """
        if select_list is None:
            select_list = []
        if group_by_clause is None:
            group_by_clause = []
        if order_by_clause is None:
            order_by_clause = []
        if exact_version_only is None:
            exact_version_only = []
        analyzer = Analyzer(
            tbl, select_list, where_clause=where_clause, group_by_clause=group_by_clause,
            order_by_clause=order_by_clause)
        row_builder = exprs.RowBuilder(analyzer.all_exprs, [], [], analyzer.sql_exprs)

        analyzer.finalize(row_builder)
        # select_list: we need to materialize everything that's been collected
        # with_pk: for now, we always retrieve the PK, because we need it for the file cache
        plan = cls._create_query_plan(
            tbl, row_builder, analyzer=analyzer, limit=limit, with_pk=True, exact_version_only=exact_version_only)
        plan.ctx.ignore_errors = ignore_errors
        select_list.clear()
        select_list.extend(analyzer.select_list)
        return plan

    @classmethod
    def _create_query_plan(
            cls, tbl: catalog.TableVersionPath, row_builder: exprs.RowBuilder, analyzer: Analyzer,
            limit: Optional[int] = None, with_pk: bool = False, exact_version_only: Optional[List[catalog.TableVersion]] = None
    ) -> exec.ExecNode:
        """
        Args:
            plan_target: if not None, generate a plan that materializes only expression that can be evaluted
                in the context of that table version (eg, if 'tbl' is a view, 'plan_target' might be the base)
        TODO: make exact_version_only a flag and use the versions from tbl
        """
        if exact_version_only is None:
            exact_version_only = []
        assert isinstance(tbl, catalog.TableVersionPath)
        is_agg_query = len(analyzer.group_by_clause) > 0 or len(analyzer.agg_fn_calls) > 0
        ctx = exec.ExecContext(row_builder)

        order_by_items = cls._determine_ordering(analyzer)
        sql_limit = 0 if is_agg_query else limit  # if we're aggregating, the limit applies to the agg output
        sql_select_list = analyzer.sql_exprs.copy()
        plan = exec.SqlScanNode(
            tbl, row_builder, select_list=sql_select_list, where_clause=analyzer.sql_where_clause,
            filter=analyzer.filter, similarity_clause=analyzer.similarity_clause, order_by_items=order_by_items,
            limit=sql_limit, set_pk=with_pk, exact_version_only=exact_version_only)
        plan = cls._insert_prefetch_node(tbl.tbl_version.id, analyzer.select_list, row_builder, plan)

        if len(analyzer.group_by_clause) > 0 or len(analyzer.agg_fn_calls) > 0:
            # we're doing aggregation; the input of the AggregateNode are the grouping exprs plus the
            # args of the agg fn calls
            agg_input = exprs.ExprSet(analyzer.group_by_clause.copy())
            for fn_call in analyzer.agg_fn_calls:
                agg_input.extend(fn_call.components)
            if not cls._is_contained_in(agg_input, analyzer.sql_exprs):
                # we need an ExprEvalNode
                plan = exec.ExprEvalNode(row_builder, agg_input, analyzer.sql_exprs, input=plan)

            # batch size for aggregation input: this could be the entire table, so we need to divide it into
            # smaller batches; at the same time, we need to make the batches large enough to amortize the
            # function call overhead
            # TODO: increase this if we have NOS calls in order to reduce the cost of switching models, but take
            # into account the amount of memory needed for intermediate images
            ctx.batch_size = 16

            plan = exec.AggregationNode(
                tbl.tbl_version, row_builder, analyzer.group_by_clause, analyzer.agg_fn_calls, agg_input, input=plan)
            agg_output = analyzer.group_by_clause + analyzer.agg_fn_calls
            if not cls._is_contained_in(analyzer.select_list, agg_output):
                # we need an ExprEvalNode to evaluate the remaining output exprs
                plan = exec.ExprEvalNode(
                    row_builder, analyzer.select_list, agg_output, input=plan)
        else:
            if not cls._is_contained_in(analyzer.select_list, analyzer.sql_exprs):
                # we need an ExprEvalNode to evaluate the remaining output exprs
                plan = exec.ExprEvalNode(row_builder, analyzer.select_list, analyzer.sql_exprs, input=plan)
            # we're returning everything to the user, so we might as well do it in a single batch
            ctx.batch_size = 0

        plan.set_ctx(ctx)
        return plan

    @classmethod
    def analyze(cls, tbl: catalog.TableVersionPath, where_clause: exprs.Predicate) -> Analyzer:
        return Analyzer(tbl, [], where_clause=where_clause)

    @classmethod
    def create_add_column_plan(
            cls, tbl: catalog.TableVersionPath, col: catalog.Column
    ) -> Tuple[exec.ExecNode, Optional[int], Optional[int]]:
        """Creates a plan for InsertableTable.add_column()
        Returns:
            plan: the plan to execute
            ctx: the context to use for the plan
            value_expr slot idx for the plan output (for computed cols)
            embedding slot idx for the plan output (for indexed image cols)
        """
        assert isinstance(tbl, catalog.TableVersionPath)
        index_info: List[Tuple[catalog.Column, func.Function]] = []
        if col.is_indexed:
            from pixeltable.functions.nos.image_embedding import openai_clip
            index_info = [(col, openai_clip)]
        row_builder = exprs.RowBuilder(
            output_exprs=[], columns=[col], indices=index_info, input_exprs=[])
        analyzer = Analyzer(tbl, row_builder.default_eval_ctx.target_exprs)
        plan = cls._create_query_plan(tbl, row_builder=row_builder, analyzer=analyzer, with_pk=True)
        plan.ctx.batch_size = 16
        plan.ctx.show_pbar = True
        plan.ctx.ignore_errors = True

        # we want to flush images
        if col.is_computed and col.is_stored and col.col_type.is_image_type():
            plan.set_stored_img_cols(row_builder.output_slot_idxs())
        value_expr_slot_idx: Optional[int] = row_builder.output_slot_idxs()[0].slot_idx if col.is_computed else None
        embedding_slot_idx: Optional[int] = row_builder.index_slot_idxs()[0].slot_idx if col.is_indexed else None
        return plan, value_expr_slot_idx, embedding_slot_idx



from .catalog import Column, Table, InsertableTable, View
from .client import Client
from .dataframe import DataFrame
from .exceptions import Error, Error
from .exprs import RELATIVE_PATH_ROOT
from .func import Function, udf, uda, Aggregator, expr_udf
from .type_system import \
    ColumnType, StringType, IntType, FloatType, BoolType, TimestampType, JsonType, ArrayType, ImageType, VideoType, \
    AudioType, DocumentType
from .utils.help import help
# noinspection PyUnresolvedReferences
from . import functions

__all__ = [
    'Client',
    'DataFrame',
    'Column',
    'Table',
    'InsertableTable',
    'View',
    'Error',
    'ColumnType',
    'StringType',
    'IntType',
    'FloatType',
    'BoolType',
    'TimestampType',
    'JsonType',
    'RELATIVE_PATH_ROOT',
    'ArrayType',
    'ImageType',
    'VideoType',
    'AudioType',
    'DocumentType',
    'Function',
    'help',
    'udf',
    'Aggregator',
    'uda',
    'expr_udf',
]






from pixeltable.type_system import \
    ColumnType, StringType, IntType, BoolType, ImageType, InvalidType, FloatType, TimestampType, JsonType, ArrayType


class TestTypes:
    def test_serialize(self, init_env) -> None:
        type_vals = [
            InvalidType(), StringType(), IntType(), BoolType(), TimestampType(),
            ImageType(height=100, width=200, mode='RGB'),
            JsonType({
                'a': StringType(), 'b': IntType(), 'c': FloatType(), 'd': BoolType(), 'e': TimestampType(),
                'f': ImageType(height=100, width=200, mode='RGB'),
                'g': JsonType({'f1': StringType(), 'f2': IntType()}),
                'h': ArrayType((224, 224, 3), dtype=IntType()),
            }),
            ArrayType((224, 224, 3), dtype=IntType()),
        ]

        for t in type_vals:
            t_serialized = t.serialize()
            t_deserialized = ColumnType.deserialize(t_serialized)
            assert t == t_deserialized



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple, Union

import jmespath
import sqlalchemy as sql

from .globals import print_slice
from .expr import Expr
from .json_mapper import JsonMapper
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable
import pixeltable.exceptions as excs
import pixeltable.catalog as catalog
import pixeltable.type_system as ts


class JsonPath(Expr):
    def __init__(self, anchor: Optional['pixeltable.exprs.ColumnRef'], path_elements: Optional[List[str]] = None, scope_idx: int = 0):
        """
        anchor can be None, in which case this is a relative JsonPath and the anchor is set later via set_anchor().
        scope_idx: for relative paths, index of referenced JsonMapper
        (0: indicates the immediately preceding JsonMapper, -1: the parent of the immediately preceding mapper, ...)
        """
        if path_elements is None:
            path_elements = []
        super().__init__(ts.JsonType())
        if anchor is not None:
            self.components = [anchor]
        self.path_elements: List[Union[str, int]] = path_elements
        self.compiled_path = jmespath.compile(self._json_path()) if len(path_elements) > 0 else None
        self.scope_idx = scope_idx
        # NOTE: the _create_id() result will change if set_anchor() gets called;
        # this is not a problem, because _create_id() shouldn't be called after init()
        self.id = self._create_id()

    def __str__(self) -> str:
        # else "R": the anchor is RELATIVE_PATH_ROOT
        return (f'{str(self._anchor) if self._anchor is not None else "R"}'
            f'{"." if isinstance(self.path_elements[0], str) else ""}{self._json_path()}')

    def _as_dict(self) -> Dict:
        return {'path_elements': self.path_elements, 'scope_idx': self.scope_idx, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'path_elements' in d
        assert 'scope_idx' in d
        assert len(components) <= 1
        anchor = components[0] if len(components) == 1 else None
        return cls(anchor, d['path_elements'], d['scope_idx'])

    @property
    def _anchor(self) -> Optional[Expr]:
        return None if len(self.components) == 0 else self.components[0]

    def set_anchor(self, anchor: Expr) -> None:
        assert len(self.components) == 0
        self.components = [anchor]

    def is_relative_path(self) -> bool:
        return self._anchor is None

    def bind_rel_paths(self, mapper: Optional['JsonMapper'] = None) -> None:
        if not self.is_relative_path():
            return
        # TODO: take scope_idx into account
        self.set_anchor(mapper.scope_anchor)

    def __call__(self, *args: object, **kwargs: object) -> 'JsonPath':
        """
        Construct a relative path that references an ancestor of the immediately enclosing JsonMapper.
        """
        if not self.is_relative_path():
            raise excs.Error(f'() for an absolute path is invalid')
        if len(args) != 1 or not isinstance(args[0], int) or args[0] >= 0:
            raise excs.Error(f'R() requires a negative index')
        return JsonPath(None, [], args[0])

    def __getattr__(self, name: str) -> 'JsonPath':
        assert isinstance(name, str)
        return JsonPath(self._anchor, self.path_elements + [name])

    def __getitem__(self, index: object) -> 'JsonPath':
        if isinstance(index, str):
            if index != '*':
                raise excs.Error(f'Invalid json list index: {index}')
        else:
            if not isinstance(index, slice) and not isinstance(index, int):
                raise excs.Error(f'Invalid json list index: {index}')
        return JsonPath(self._anchor, self.path_elements + [index])

    def __rshift__(self, other: object) -> 'JsonMapper':
        rhs_expr = Expr.from_object(other)
        if rhs_expr is None:
            raise excs.Error(f'>> requires an expression on the right-hand side, found {type(other)}')
        return JsonMapper(self, rhs_expr)

    def default_column_name(self) -> Optional[str]:
        anchor_name = self._anchor.default_column_name() if self._anchor is not None else ''
        ret_name = f'{anchor_name}.{self._json_path()}'
        
        def cleanup_char(s : str) -> str:
            if s == '.':
                return '_'
            elif s == '*':
                return 'star'
            elif s.isalnum():
                return s
            else:
                return ''
            
        clean_name = ''.join(map(cleanup_char, ret_name))
        clean_name = clean_name.lstrip('_') # remove leading underscore
        if clean_name == '':
            clean_name = None
        
        assert clean_name is None or catalog.is_valid_identifier(clean_name)
        return clean_name

    def _equals(self, other: JsonPath) -> bool:
        return self.path_elements == other.path_elements

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('path_elements', self.path_elements)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        """
        Postgres appears to have a bug: jsonb_path_query('{a: [{b: 0}, {b: 1}]}', '$.a.b') returns
        *two* rows (each containing col val 0), not a single row with [0, 0].
        We need to use a workaround: retrieve the entire dict, then use jmespath to extract the path correctly.
        """
        #path_str = '$.' + '.'.join(self.path_elements)
        #assert isinstance(self._anchor(), ColumnRef)
        #return sql.func.jsonb_path_query(self._anchor().col.sa_col, path_str)
        return None

    def _json_path(self) -> str:
        assert len(self.path_elements) > 0
        result: List[str] = []
        for element in self.path_elements:
            if element == '*':
                result.append('[*]')
            elif isinstance(element, str):
                result.append(f'{"." if len(result) > 0 else ""}{element}')
            elif isinstance(element, int):
                result.append(f'[{element}]')
            elif isinstance(element, slice):
                result.append(f'[{print_slice(element)}]')
        return ''.join(result)

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        val = data_row[self._anchor.slot_idx]
        if self.compiled_path is not None:
            val = self.compiled_path.search(val)
        data_row[self.slot_idx] = val


RELATIVE_PATH_ROOT = JsonPath(None)



import pytest

import pixeltable as pxt
from pixeltable import exceptions as excs
from pixeltable.tests.utils import make_tbl


class TestDirs:
    def test_create(self, test_client: pxt.Client) -> None:
        cl = test_client
        dirs = ['dir1', 'dir1.sub1', 'dir1.sub1.subsub1']
        for name in dirs:
            cl.create_dir(name)

        # invalid names
        with pytest.raises(excs.Error):
            cl.create_dir('1dir')
        with pytest.raises(excs.Error):
            cl.create_dir('_dir1')
        with pytest.raises(excs.Error):
            cl.create_dir('dir 1')
        with pytest.raises(excs.Error):
            cl.create_dir('dir1..sub2')
        with pytest.raises(excs.Error):
            cl.create_dir('dir1.sub2.')
        with pytest.raises(excs.Error):
            cl.create_dir('dir1:sub2.')

        # existing dirs
        with pytest.raises(excs.Error):
            cl.create_dir('dir1')
        cl.create_dir('dir1', ignore_errors=True)
        with pytest.raises(excs.Error):
            cl.create_dir('dir1.sub1')
        with pytest.raises(excs.Error):
            cl.create_dir('dir1.sub1.subsub1')

        # existing table
        make_tbl(cl, 'dir1.t1')
        with pytest.raises(excs.Error):
            cl.create_dir('dir1.t1')

        with pytest.raises(excs.Error):
            cl.create_dir('dir2.sub2')
        make_tbl(cl, 't2')
        with pytest.raises(excs.Error):
            cl.create_dir('t2.sub2')

        # new client: force loading from store
        cl2 = pxt.Client(reload=True)

        listing = cl2.list_dirs(recursive=True)
        assert listing == dirs
        listing = cl2.list_dirs(recursive=False)
        assert listing == ['dir1']
        listing = cl2.list_dirs('dir1', recursive=True)
        assert listing == ['dir1.sub1', 'dir1.sub1.subsub1']
        listing = cl2.list_dirs('dir1', recursive=False)
        assert listing == ['dir1.sub1']
        listing = cl2.list_dirs('dir1.sub1', recursive=True)
        assert listing == ['dir1.sub1.subsub1']
        listing = cl2.list_dirs('dir1.sub1', recursive=False)
        assert listing == ['dir1.sub1.subsub1']

    def test_rm(self, test_client: pxt.Client) -> None:
        cl = test_client
        dirs = ['dir1', 'dir1.sub1', 'dir1.sub1.subsub1']
        for name in dirs:
            cl.create_dir(name)
        make_tbl(cl, 't1')
        make_tbl(cl, 'dir1.t1')

        # bad name
        with pytest.raises(excs.Error):
            cl.rm_dir('1dir')
        # bad path
        with pytest.raises(excs.Error):
            cl.rm_dir('dir1..sub1')
        # doesn't exist
        with pytest.raises(excs.Error):
            cl.rm_dir('dir2')
        # not empty
        with pytest.raises(excs.Error):
            cl.rm_dir('dir1')

        cl.rm_dir('dir1.sub1.subsub1')
        assert cl.list_dirs('dir1.sub1') == []

        # check after reloading
        cl = pxt.Client(reload=True)
        assert cl.list_dirs('dir1.sub1') == []

    def test_move(self, test_client: pxt.Client) -> None:
        cl = test_client
        cl.create_dir('dir1')
        cl.create_dir('dir1.sub1')
        make_tbl(cl, 'dir1.sub1.t1')
        assert cl.list_tables('dir1') == ['dir1.sub1.t1']
        cl.move('dir1.sub1.t1', 'dir1.sub1.t2')
        assert cl.list_tables('dir1') == ['dir1.sub1.t2']
        cl.create_dir('dir2')
        cl.move('dir1', 'dir2.dir1')
        assert cl.list_tables('dir2') == ['dir2.dir1.sub1.t2']

        # new client: force loading from store
        cl2 = pxt.Client(reload=True)
        assert cl2.list_tables('dir2') == ['dir2.dir1.sub1.t2']



from __future__ import annotations

import base64
import copy
import hashlib
import io
import json
import logging
import mimetypes
import traceback
from pathlib import Path
from typing import List, Optional, Any, Dict, Generator, Tuple, Set

import pandas as pd
import pandas.io.formats.style
import sqlalchemy as sql
from PIL import Image

import pixeltable.catalog as catalog
import pixeltable.exceptions as excs
import pixeltable.exprs as exprs
import pixeltable.type_system as ts
from pixeltable.catalog import is_valid_identifier
from pixeltable.env import Env
from pixeltable.plan import Planner
from pixeltable.type_system import ColumnType

__all__ = [
    'DataFrame'
]

_logger = logging.getLogger('pixeltable')

def _format_img(img: object) -> str:
    """
    Create <img> tag for Image object.
    """
    assert isinstance(img, Image.Image), f'Wrong type: {type(img)}'
    with io.BytesIO() as buffer:
        img.save(buffer, 'jpeg')
        img_base64 = base64.b64encode(buffer.getvalue()).decode()
        return f'<div style=\'width:200px;\'><img src="data:image/jpeg;base64,{img_base64}"></div>'

def _create_source_tag(file_path: str) -> str:
    abs_path = Path(file_path)
    assert abs_path.is_absolute()
    src_url = f'{Env.get().http_address}/{abs_path}'
    mime = mimetypes.guess_type(src_url)[0]
    # if mime is None, the attribute string would not be valid html.
    mime_attr = f'type="{mime}"' if mime is not None else ''
    return f'<source src="{src_url}" {mime_attr} />'

def _format_video(file_path: str) -> str:
    return f'<video controls>{_create_source_tag(file_path)}</video>'

def _format_audio(file_path: str) -> str:
    return f'<audio controls>{_create_source_tag(file_path)}</audio>'

class DataFrameResultSet:
    def __init__(self, rows: List[List[Any]], col_names: List[str], col_types: List[ColumnType]):
        self._rows = rows
        self._col_names = col_names
        self._col_types = col_types
        self._formatters = {
            ts.ImageType: _format_img,
            ts.VideoType: _format_video,
            ts.AudioType: _format_audio,
        }

    def __len__(self) -> int:
        return len(self._rows)

    def column_names(self) -> List[str]:
        return self._col_names

    def column_types(self) -> List[ColumnType]:
        return self._col_types

    def __repr__(self) -> str:
        return self.to_pandas().__repr__()

    def _repr_html_(self) -> str:
        formatters = {
            col_name: self._formatters[col_type.__class__]
            for col_name, col_type in zip(self._col_names, self._col_types)
            if col_type.__class__ in self._formatters
        }

        # TODO: why does mypy complain about formatters having an incorrect type?
        return self.to_pandas().to_html(formatters=formatters, escape=False, index=False)  # type: ignore[arg-type]

    def __str__(self) -> str:
        return self.to_pandas().to_string()

    def _reverse(self) -> None:
        """Reverse order of rows"""
        self._rows.reverse()

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame.from_records(self._rows, columns=self._col_names)

    def _row_to_dict(self, row_idx: int) -> Dict[str, Any]:
        return {self._col_names[i]: self._rows[row_idx][i] for i in range(len(self._col_names))}

    def __getitem__(self, index: Any) -> Any:
        if isinstance(index, str):
            if index not in self._col_names:
                raise excs.Error(f'Invalid column name: {index}')
            col_idx = self._col_names.index(index)
            return [row[col_idx] for row in self._rows]
        if isinstance(index, int):
            return self._row_to_dict(index)
        if isinstance(index, tuple) and len(index) == 2:
            if not isinstance(index[0], int) or not (isinstance(index[1], str) or isinstance(index[1], int)):
                raise excs.Error(f'Bad index, expected [<row idx>, <column name | column index>]: {index}')
            if isinstance(index[1], str) and index[1] not in self._col_names:
                raise excs.Error(f'Invalid column name: {index[1]}')
            col_idx = self._col_names.index(index[1]) if isinstance(index[1], str) else index[1]
            return self._rows[index[0]][col_idx]
        raise excs.Error(f'Bad index: {index}')

    def __iter__(self) -> DataFrameResultSetIterator:
        return DataFrameResultSetIterator(self)

    def __eq__(self, other):
        if not isinstance(other, DataFrameResultSet):
            return False
        return self.to_pandas().equals(other.to_pandas())


class DataFrameResultSetIterator:
    def __init__(self, result_set: DataFrameResultSet):
        self._result_set = result_set
        self._idx = 0

    def __next__(self) -> Dict[str, Any]:
        if self._idx >= len(self._result_set):
            raise StopIteration
        row = self._result_set._row_to_dict(self._idx)
        self._idx += 1
        return row


# TODO: remove this; it's only here as a reminder that we still need to call release() in the current implementation
class AnalysisInfo:
    def __init__(self, tbl: catalog.TableVersion):
        self.tbl = tbl
        # output of the SQL scan stage
        self.sql_scan_output_exprs: List[exprs.Expr] = []
        # output of the agg stage
        self.agg_output_exprs: List[exprs.Expr] = []
        # Where clause of the Select stmt of the SQL scan stage
        self.sql_where_clause: Optional[sql.ClauseElement] = None
        # filter predicate applied to input rows of the SQL scan stage
        self.filter: Optional[exprs.Predicate] = None
        self.similarity_clause: Optional[exprs.ImageSimilarityPredicate] = None
        self.agg_fn_calls: List[exprs.FunctionCall] = []  # derived from unique_exprs
        self.has_frame_col: bool = False  # True if we're referencing the frame col

        self.evaluator: Optional[exprs.Evaluator] = None
        self.sql_scan_eval_ctx: List[exprs.Expr] = []  # needed to materialize output of SQL scan stage
        self.agg_eval_ctx: List[exprs.Expr] = []  # needed to materialize output of agg stage
        self.filter_eval_ctx: List[exprs.Expr] = []
        self.group_by_eval_ctx: List[exprs.Expr] = []

    def finalize_exec(self) -> None:
        """
        Call release() on all collected Exprs.
        """
        exprs.Expr.release_list(self.sql_scan_output_exprs)
        exprs.Expr.release_list(self.agg_output_exprs)
        if self.filter is not None:
            self.filter.release()



class DataFrame:
    def __init__(
            self, tbl: catalog.TableVersionPath,
            select_list: Optional[List[Tuple[exprs.Expr, Optional[str]]]] = None,
            where_clause: Optional[exprs.Predicate] = None,
            group_by_clause: Optional[List[exprs.Expr]] = None,
            grouping_tbl: Optional[catalog.TableVersion] = None,
            order_by_clause: Optional[List[Tuple[exprs.Expr, bool]]] = None,  # List[(expr, asc)]
            limit: Optional[int] = None):
        self.tbl = tbl

        # select list logic
        DataFrame._select_list_check_rep(select_list) # check select list without expansion
        # exprs contain execution state and therefore cannot be shared
        select_list = copy.deepcopy(select_list)
        select_list_exprs, column_names = DataFrame._normalize_select_list(tbl, select_list)
        DataFrame._select_list_check_rep(list(zip(select_list_exprs, column_names)))
        # check select list after expansion to catch early
        # the following two lists are always non empty, even if select list is None.
        self._select_list_exprs = select_list_exprs
        self._column_names = column_names
        self.select_list = select_list

        self.where_clause = copy.deepcopy(where_clause)
        assert group_by_clause is None or grouping_tbl is None
        self.group_by_clause = copy.deepcopy(group_by_clause)
        self.grouping_tbl = grouping_tbl
        self.order_by_clause = copy.deepcopy(order_by_clause)
        self.limit_val = limit

    @classmethod
    def _select_list_check_rep(cls,
        select_list: Optional[List[Tuple[exprs.Expr, Optional[str]]]],
    ) -> None:
        """Validate basic select list types. 
        """
        if select_list is None: # basic check for valid select list
            return

        assert len(select_list) > 0
        for ent in select_list:
            assert isinstance(ent, tuple)
            assert len(ent) == 2
            assert isinstance(ent[0], exprs.Expr)
            assert ent[1] is None or isinstance(ent[1], str)
            if isinstance(ent[1], str):
                assert is_valid_identifier(ent[1])

    @classmethod
    def _normalize_select_list(cls,
        tbl: catalog.TableVersionPath,
        select_list: Optional[List[Tuple[exprs.Expr, Optional[str]]]],
    ) -> Tuple[List[exprs.Expr], List[str]]:
        """
        Expand select list information with all columns and their names
        Returns: 
            a pair composed of the list of expressions and the list of corresponding names
        """
        if select_list is None:
            expanded_list = [(exprs.ColumnRef(col), None) for col in tbl.columns()]
        else:
            expanded_list = select_list

        out_exprs : List[exprs.Expr] = []
        out_names : List[str] = [] # keep track of order
        seen_out_names : set[str] = set() # use to check for duplicates in loop, avoid square complexity
        for i, (expr, name) in enumerate(expanded_list):
            if name is None:
                # use default, add suffix if needed so default adds no duplicates
                default_name = expr.default_column_name()
                if default_name is not None:
                    column_name = default_name
                    if default_name in seen_out_names:
                        # already used, then add suffix until unique name is found
                        for j in range(1, len(out_names)+1):
                            column_name = f'{default_name}_{j}'
                            if column_name not in seen_out_names:
                                break
                else: # no default name, eg some expressions
                    column_name = f'col_{i}'
            else: # user provided name, no attempt to rename
                column_name = name

            out_exprs.append(expr)
            out_names.append(column_name)
            seen_out_names.add(column_name)
        assert len(out_exprs) == len(out_names)
        assert set(out_names) == seen_out_names
        return out_exprs, out_names

    def _exec(self) -> Generator[exprs.DataRow, None, None]:
        """Run the query and return rows as a generator.
        This function must not modify the state of the DataFrame, otherwise it breaks dataset caching.
        """
        # construct a group-by clause if we're grouping by a table
        group_by_clause: List[exprs.Expr] = []
        if self.grouping_tbl is not None:
            assert self.group_by_clause is None
            num_rowid_cols = len(self.grouping_tbl.store_tbl.rowid_columns())
            # the grouping table must be a base of self.tbl
            assert num_rowid_cols <= len(self.tbl.tbl_version.store_tbl.rowid_columns())
            group_by_clause = [exprs.RowidRef(self.tbl.tbl_version, idx) for idx in range(num_rowid_cols)]
        elif self.group_by_clause is not None:
            group_by_clause = self.group_by_clause

        for item in self._select_list_exprs:
            item.bind_rel_paths(None)
        plan = Planner.create_query_plan(
            self.tbl, self._select_list_exprs, where_clause=self.where_clause, group_by_clause=group_by_clause,
            order_by_clause=self.order_by_clause if self.order_by_clause is not None else [],
            limit=self.limit_val if self.limit_val is not None else 0)  # limit_val == 0: no limit_val

        with Env.get().engine.begin() as conn:
            plan.ctx.conn = conn
            plan.open()
            try:
                for row_batch in plan:
                    for data_row in row_batch:
                        yield data_row
            finally:
                plan.close()
            return

    def show(self, n: int = 20) -> DataFrameResultSet:
        assert n is not None
        return self.limit(n).collect()

    def head(self, n: int = 10) -> DataFrameResultSet:
        if self.order_by_clause is not None:
            raise excs.Error(f'head() cannot be used with order_by()')
        num_rowid_cols = len(self.tbl.tbl_version.store_tbl.rowid_columns())
        order_by_clause = [exprs.RowidRef(self.tbl.tbl_version, idx) for idx in range(num_rowid_cols)]
        return self.order_by(*order_by_clause, asc=True).limit(n).collect()

    def tail(self, n: int = 10) -> DataFrameResultSet:
        if self.order_by_clause is not None:
            raise excs.Error(f'tail() cannot be used with order_by()')
        num_rowid_cols = len(self.tbl.tbl_version.store_tbl.rowid_columns())
        order_by_clause = [exprs.RowidRef(self.tbl.tbl_version, idx) for idx in range(num_rowid_cols)]
        result = self.order_by(*order_by_clause, asc=False).limit(n).collect()
        result._reverse()
        return result

    def get_column_names(self) -> List[str]:
        return self._column_names

    def get_column_types(self) -> List[ColumnType]:
        return [expr.col_type for expr in self._select_list_exprs]

    def collect(self) -> DataFrameResultSet:
        try:
            result_rows = []
            for data_row in self._exec():
                result_row = [data_row[e.slot_idx] for e in self._select_list_exprs]
                result_rows.append(result_row)
        except excs.ExprEvalError as e:
            msg = (f'In row {e.row_num} the {e.expr_msg} encountered exception '
                   f'{type(e.exc).__name__}:\n{str(e.exc)}')
            if len(e.input_vals) > 0:
                input_msgs = [
                    f"'{d}' = {d.col_type.print_value(e.input_vals[i])}"
                    for i, d in enumerate(e.expr.dependencies())
                ]
                msg += f'\nwith {", ".join(input_msgs)}'
            assert e.exc_tb is not None
            stack_trace = traceback.format_tb(e.exc_tb)
            if len(stack_trace) > 2:
                # append a stack trace if the exception happened in user code
                # (frame 0 is ExprEvaluator and frame 1 is some expr's eval()
                nl = '\n'
                # [-1:0:-1]: leave out entry 0 and reverse order, so that the most recent frame is at the top
                msg += f'\nStack:\n{nl.join(stack_trace[-1:1:-1])}'
            raise excs.Error(msg)
        except sql.exc.DBAPIError as e:
            raise excs.Error(f'Error during SQL execution:\n{e}')

        col_types = self.get_column_types()
        return DataFrameResultSet(result_rows, self._column_names, col_types)

    def count(self) -> int:
        from pixeltable.plan import Planner
        stmt = Planner.create_count_stmt(self.tbl, self.where_clause)
        with Env.get().engine.connect() as conn:
            result: int = conn.execute(stmt).scalar_one()
            assert isinstance(result, int)
            return result

    def _description(self) -> pd.DataFrame:
        """see DataFrame.describe()"""
        heading_vals: List[str] = []
        info_vals: List[str] = []
        if self.select_list is not None:
            assert len(self.select_list) > 0
            heading_vals.append('Select')
            heading_vals.extend([''] * (len(self.select_list) - 1))
            info_vals.extend(self.get_column_names())
        if self.where_clause is not None:
            heading_vals.append('Where')
            info_vals.append(self.where_clause.display_str(inline=False))
        if self.group_by_clause is not None:
            heading_vals.append('Group By')
            heading_vals.extend([''] * (len(self.group_by_clause) - 1))
            info_vals.extend([e.display_str(inline=False) for e in self.group_by_clause])
        if self.order_by_clause is not None:
            heading_vals.append('Order By')
            heading_vals.extend([''] * (len(self.order_by_clause) - 1))
            info_vals.extend([
                f'{e[0].display_str(inline=False)} {"asc" if e[1] else "desc"}' for e in self.order_by_clause
            ])
        if self.limit_val is not None:
            heading_vals.append('Limit')
            info_vals.append(str(self.limit_val))
        assert len(heading_vals) > 0
        assert len(info_vals) > 0
        assert len(heading_vals) == len(info_vals)
        return pd.DataFrame({'Heading': heading_vals, 'Info': info_vals})

    def _description_html(self) -> pandas.io.formats.style.Styler:
        """Return the description in an ipython-friendly manner."""
        pd_df = self._description()
        # white-space: pre-wrap: print \n as newline
        # th: center-align headings
        return pd_df.style.set_properties(**{'white-space': 'pre-wrap', 'text-align': 'left'}) \
            .set_table_styles([dict(selector='th', props=[('text-align', 'center')])]) \
            .hide(axis='index').hide(axis='columns')

    def describe(self) -> None:
        """
        Prints a tabular description of this DataFrame.
        The description has two columns, heading and info, which list the contents of each 'component'
                (select list, where clause, ...) vertically.
        """
        try:
            __IPYTHON__
            from IPython.display import display
            display(self._description_html())
        except NameError:
            print(self.__repr__())

    def __repr__(self) -> str:
        return self._description().to_string(header=False, index=False)

    def _repr_html_(self) -> str:
        return self._description_html()._repr_html_()

    def select(self, *items: Any, **named_items : Any) -> DataFrame:
        if self.select_list is not None:
            raise excs.Error(f'Select list already specified')
        for (name, _) in named_items.items():
            if not isinstance(name, str) or not is_valid_identifier(name):
                raise excs.Error(f'Invalid name: {name}')
        base_list = [(expr, None) for expr in items] + [(expr, k) for (k, expr) in named_items.items()]
        if len(base_list) == 0:
            raise excs.Error(f'Empty select list')
        
        # analyze select list; wrap literals with the corresponding expressions
        select_list = []
        for raw_expr, name in base_list:
            if isinstance(raw_expr, exprs.Expr):
                select_list.append((raw_expr, name))
            elif isinstance(raw_expr, dict):
                select_list.append((exprs.InlineDict(raw_expr), name))
            elif isinstance(raw_expr, list):
                select_list.append((exprs.InlineArray(raw_expr), name))
            else:
                select_list.append((exprs.Literal(raw_expr), name))
            expr = select_list[-1][0]
            if expr.col_type.is_invalid_type():
                raise excs.Error(f'Invalid type: {raw_expr}')
            # TODO: check that ColumnRefs in expr refer to self.tbl

        # check user provided names do not conflict among themselves
        # or with auto-generated ones
        seen: Set[str] = set()
        _, names = DataFrame._normalize_select_list(self.tbl, select_list)
        for name in names:
            if name in seen:
                repeated_names = [j for j, x in enumerate(names) if x == name]
                pretty = ', '.join(map(str, repeated_names))
                raise excs.Error(f'Repeated column name "{name}" in select() at positions: {pretty}')
            seen.add(name)

        return DataFrame(
            self.tbl, select_list=select_list, where_clause=self.where_clause, group_by_clause=self.group_by_clause,
            grouping_tbl=self.grouping_tbl, order_by_clause=self.order_by_clause, limit=self.limit_val)

    def where(self, pred: exprs.Predicate) -> DataFrame:
        return DataFrame(
            self.tbl, select_list=self.select_list, where_clause=pred, group_by_clause=self.group_by_clause,
            grouping_tbl=self.grouping_tbl, order_by_clause=self.order_by_clause, limit=self.limit_val)

    def group_by(self, *grouping_items: Any) -> DataFrame:
        """Add a group-by clause to this DataFrame.
        Variants:
        - group_by(<base table>): group a component view by their respective base table rows
        - group_by(<expr>, ...): group by the given expressions
        """
        if self.group_by_clause is not None:
            raise excs.Error(f'Group-by already specified')
        grouping_tbl: Optional[catalog.TableVersion] = None
        group_by_clause: Optional[List[exprs.Expr]] = None
        for item in grouping_items:
            if isinstance(item, catalog.Table):
                if len(grouping_items) > 1:
                    raise excs.Error(f'group_by(): only one table can be specified')
                # we need to make sure that the grouping table is a base of self.tbl
                base = self.tbl.find_tbl_version(item.tbl_version_path.tbl_id())
                if base is None or base.id == self.tbl.tbl_id():
                    raise excs.Error(f'group_by(): {item.name} is not a base table of {self.tbl.tbl_name()}')
                grouping_tbl = item.tbl_version_path.tbl_version
                break
            if not isinstance(item, exprs.Expr):
                raise excs.Error(f'Invalid expression in group_by(): {item}')
        if grouping_tbl is None:
            group_by_clause = list(grouping_items)
        return DataFrame(
            self.tbl, select_list=self.select_list, where_clause=self.where_clause, group_by_clause=group_by_clause,
            grouping_tbl=grouping_tbl, order_by_clause=self.order_by_clause, limit=self.limit_val)

    def order_by(self, *expr_list: exprs.Expr, asc: bool = True) -> DataFrame:
        for e in expr_list:
            if not isinstance(e, exprs.Expr):
                raise excs.Error(f'Invalid expression in order_by(): {e}')
        order_by_clause = self.order_by_clause if self.order_by_clause is not None else []
        order_by_clause.extend([(e.copy(), asc) for e in expr_list])
        return DataFrame(
            self.tbl, select_list=self.select_list, where_clause=self.where_clause,
            group_by_clause=self.group_by_clause, grouping_tbl=self.grouping_tbl, order_by_clause=order_by_clause,
            limit=self.limit_val)

    def limit(self, n: int) -> DataFrame:
        assert n is not None and isinstance(n, int)
        return DataFrame(
            self.tbl, select_list=self.select_list, where_clause=self.where_clause,
            group_by_clause=self.group_by_clause, grouping_tbl=self.grouping_tbl, order_by_clause=self.order_by_clause,
            limit=n)

    def __getitem__(self, index: object) -> DataFrame:
        """
        Allowed:
        - [<Predicate>]: filter operation
        - [List[Expr]]/[Tuple[Expr]]: setting the select list
        - [Expr]: setting a single-col select list
        """
        if isinstance(index, exprs.Predicate):
            return self.where(index)
        if isinstance(index, tuple):
            index = list(index)
        if isinstance(index, exprs.Expr):
            index = [index]
        if isinstance(index, list):
            return self.select(*index)
        raise TypeError(f'Invalid index type: {type(index)}')
    
    def _as_dict(self) -> Dict[str, Any]:
        """ 
            Returns:
                Dictionary representing this dataframe.
        """
        tbl_versions = self.tbl.get_tbl_versions()
        d = {
            '_classname': 'DataFrame',
            'tbl_ids': [str(t.id) for t in tbl_versions],
            'tbl_versions': [t.version for t in tbl_versions],
            'select_list':
                [(e.as_dict(), name) for (e, name) in self.select_list] if self.select_list is not None else None,
            'where_clause': self.where_clause.as_dict() if self.where_clause is not None else None,
            'group_by_clause':
                [e.as_dict() for e in self.group_by_clause] if self.group_by_clause is not None else None,
            'order_by_clause':
                [(e.as_dict(), asc) for (e,asc) in self.order_by_clause] if self.order_by_clause is not None else None,
            'limit_val': self.limit_val,
        }
        return d

    def to_coco_dataset(self) -> Path:
        """Convert the dataframe to a COCO dataset.
        This dataframe must return a single json-typed output column in the following format:
        {
            'image': PIL.Image.Image,
            'annotations': [
                {
                    'bbox': [x: int, y: int, w: int, h: int],
                    'category': str | int,
                },
                ...
            ],
        }

        Returns:
            Path to the COCO dataset file.
        """
        from pixeltable.utils.coco import write_coco_dataset

        summary_string = json.dumps(self._as_dict())
        cache_key = hashlib.sha256(summary_string.encode()).hexdigest()

        dest_path = (Env.get().dataset_cache_dir / f'coco_{cache_key}')
        if dest_path.exists():
            assert dest_path.is_dir()
            data_file_path = dest_path / 'data.json'
            assert data_file_path.exists()
            assert data_file_path.is_file()
            return data_file_path
        else:
            return write_coco_dataset(self, dest_path)

    # TODO Factor this out into a separate module.
    # The return type is unresolvable, but torch can't be imported since it's an optional dependency.
    def to_pytorch_dataset(self, image_format: str = 'pt') -> 'torch.utils.data.IterableDataset':
        """
        Convert the dataframe to a pytorch IterableDataset suitable for parallel loading
        with torch.utils.data.DataLoader.

        This method requires pyarrow >= 13, torch and torchvision to work.

        This method serializes data so it can be read from disk efficiently and repeatedly without
        re-executing the query. This data is cached to disk for future re-use.

        Args:
            image_format: format of the images. Can be 'pt' (pytorch tensor) or 'np' (numpy array).
                    'np' means image columns return as an RGB uint8 array of shape HxWxC.
                    'pt' means image columns return as a CxHxW tensor with values in [0,1] and type torch.float32.
                        (the format output by torchvision.transforms.ToTensor())

        Returns:
            A pytorch IterableDataset: Columns become fields of the dataset, where rows are returned as a dictionary
                compatible with torch.utils.data.DataLoader default collation.

        Constraints:
            The default collate_fn for torch.data.util.DataLoader cannot represent null values as part of a
            pytorch tensor when forming batches. These values will raise an exception while running the dataloader.

            If you have them, you can work around None values by providing your custom collate_fn to the DataLoader
            (and have your model handle it). Or, if these are not meaningful values within a minibtach, you can
            modify or remove any such values through selections and filters prior to calling to_pytorch_dataset().
        """
        # check dependencies
        Env.get().require_package('pyarrow', [13])
        Env.get().require_package('torch')
        Env.get().require_package('torchvision')

        from pixeltable.utils.parquet import save_parquet # pylint: disable=import-outside-toplevel
        from pixeltable.utils.pytorch import PixeltablePytorchDataset # pylint: disable=import-outside-toplevel

        summary_string = json.dumps(self._as_dict()) 
        cache_key = hashlib.sha256(summary_string.encode()).hexdigest()
    
        dest_path = (Env.get().dataset_cache_dir / f'df_{cache_key}').with_suffix('.parquet') # pylint: disable = protected-access
        if dest_path.exists(): # fast path: use cache
            assert dest_path.is_dir()
        else:
            save_parquet(self, dest_path)

        return PixeltablePytorchDataset(path=dest_path, image_format=image_format)
